# pylint: disable=E501
"""Functions to compute the Jacobian for EFC."""
import numpy as np
from numpy.fft import fftshift, fft2

import falco
from falco.util import pad_crop
import falco.prop as fp


def lyot(mp, im, idm):
    """
    Differential model used to compute the ctrl Jacobian for Lyot coronagraph.

    Specialized compact model used to compute the DM response matrix, aka the
    control Jacobian for a Lyot coronagraph. Can include an apodizer, making it
    an apodized pupil Lyot coronagraph (APLC). Does not include unknown
    aberrations of the full, "truth" model. This model propagates the
    first-order Taylor expansion of the phase from the poke of each actuator
    of the deformable mirror.

    Parameters
    ----------
    mp : ModelParameters
        Structure containing optical model parameters

    Returns
    -------
    Gzdl : numpy ndarray
        Complex-valued, 2-D array containing the Jacobian for the
        specified Zernike mode, DM number, and wavelength.

    """
    modvar = falco.config.Object()  # Initialize the new structure
    modvar.sbpIndex = mp.jac.sbp_inds[im]
    modvar.zernIndex = mp.jac.zern_inds[im]

    wvl = mp.sbp_centers[modvar.sbpIndex]
    mirrorFac = 2.  # Phase change is twice the DM surface height.
    NdmPad = int(mp.compact.NdmPad)
    
    if mp.coro.upper() in ('LC', 'APLC', 'FLC', 'SPLC'):
        fpm = mp.F3.compact.ampMask
        transOuterFPM = 1.  # transmission of points outside the FPM.
    elif mp.coro.upper() in ('HLC',):
        fpm = np.squeeze(mp.compact.fpmCube[:, :, modvar.sbpIndex])  # complex
        # Complex transmission of the points outside the FPM (just fused silica
        # with optional dielectric and no metal).
        transOuterFPM = fpm[0, 0]

    """Input E-fields"""
    Ein = np.squeeze(mp.P1.compact.E[:, :, modvar.sbpIndex])
    # Apply a Zernike (in amplitude) at input pupil
    # Used only for Zernike sensitivity control, which requires the perfect
    # E-field of the differential Zernike term.
    if not (modvar.zernIndex == 1):
        indsZnoll = modvar.zernIndex  # Just send in 1 Zernike mode
        zernMat = np.squeeze(falco.zern.gen_norm_zern_maps(mp.P1.compact.Nbeam,
                                                    mp.centering, indsZnoll))
        zernMat = pad_crop(zernMat, mp.P1.compact.Narr)
        Ein = Ein*zernMat*(2*np.pi/wvl)*mp.jac.Zcoef[mp.jac.zerns ==
                                                     modvar.zernIndex]

    """ Masks and DM surfaces """
    pupil = pad_crop(mp.P1.compact.mask, NdmPad)
    Ein = pad_crop(Ein, NdmPad)

    # Re-image the apodizer from pupil P3 back to pupil P2.
    if(mp.flagApod):
        apodReimaged = pad_crop(mp.P3.compact.mask, NdmPad)
        apodReimaged = fp.relay(apodReimaged, mp.Nrelay2to3, mp.centering)
    else:
        apodReimaged = np.ones((NdmPad, NdmPad))

    # Compute the DM surfaces for the current DM commands
    if any(mp.dm_ind == 1):
        DM1surf = pad_crop(mp.dm1.compact.surfM, NdmPad)
        # DM1surf = falco.dm.gen_surf_from_act(mp.dm1, mp.dm1.compact.dx, NdmPad)
    else:
        DM1surf = np.zeros((NdmPad, NdmPad))
    if any(mp.dm_ind == 2):
        DM2surf = pad_crop(mp.dm2.compact.surfM, NdmPad)
        # DM2surf = falco.dm.gen_surf_from_act(mp.dm2, mp.dm2.compact.dx, NdmPad)
    else:
        DM2surf = np.zeros((NdmPad, NdmPad))

    if mp.flagDM1stop:
        DM1stop = pad_crop(mp.dm1.compact.mask, NdmPad) 
    else:
        DM1stop = np.ones((NdmPad, NdmPad))
    if(mp.flagDM2stop):
        DM2stop = pad_crop(mp.dm2.compact.mask, NdmPad) 
    else:
        DM2stop = np.ones((NdmPad, NdmPad))

    # This block is for BMC surface error testing
    if mp.flagDMwfe:  # if(mp.flagDMwfe && (mp.P1.full.Nbeam==mp.P1.compact.Nbeam))
        if any(mp.dm_ind == 1):
            Edm1WFE = np.exp(2*np.pi*1j/wvl*pad_crop(mp.dm1.compact.wfe, NdmPad, 'extrapval', 0))
        else:
            Edm1WFE = np.ones((NdmPad, NdmPad))
        if any(mp.dm_ind == 2):
            Edm2WFE = np.exp(2*np.pi*1j/wvl*pad_crop(mp.dm2.compact.wfe, NdmPad, 'extrapval', 0))
        else:
            Edm2WFE = np.ones((NdmPad, NdmPad))
    else:
        Edm1WFE = np.ones((NdmPad, NdmPad))
        Edm2WFE = np.ones((NdmPad, NdmPad))

    """Propagation"""
    # Define pupil P1 and Propagate to pupil P2
    EP1 = pupil*Ein  # E-field at pupil plane P1
    EP2 = fp.relay(EP1, mp.Nrelay1to2, mp.centering)

    # Propagate from P2 to DM1, and apply DM1 surface and aperture stop
    if not abs(mp.d_P2_dm1) == 0:
        Edm1 = fp.ptp(EP2, mp.P2.compact.dx*NdmPad, wvl, mp.d_P2_dm1)
    else:
        Edm1 = EP2
    Edm1out = Edm1*Edm1WFE*DM1stop*np.exp(mirrorFac*2*np.pi*1j*DM1surf/wvl)

    """ ---------- DM1 ---------- """
    if idm == 1:
        Gzdl = np.zeros((mp.Fend.corr.Npix, mp.dm1.Nele), dtype=np.complex)
        
        # Two array sizes (at same resolution) of influence functions for MFT
        # and angular spectrum
        NboxPad1AS = int(mp.dm1.compact.NboxAS)  # array size for FFT-AS propagations from DM1->DM2->DM1
        # Adjust the sub-array location of the influence function for the added zero padding
        mp.dm1.compact.xy_box_lowerLeft_AS = mp.dm1.compact.xy_box_lowerLeft -\
            (mp.dm1.compact.NboxAS-mp.dm1.compact.Nbox)/2.
    
        if any(mp.dm_ind == 2):
            DM2surf = pad_crop(DM2surf, mp.dm1.compact.NdmPad)
        else:
            DM2surf = np.zeros((mp.dm1.compact.NdmPad, mp.dm1.compact.NdmPad))
        if(mp.flagDM2stop):
            DM2stop = pad_crop(DM2stop, mp.dm1.compact.NdmPad)
        else:
            DM2stop = np.ones((mp.dm1.compact.NdmPad, mp.dm1.compact.NdmPad))
        apodReimaged = pad_crop(apodReimaged, mp.dm1.compact.NdmPad)
    
        Edm1pad = pad_crop(Edm1out, mp.dm1.compact.NdmPad)  # Pad or crop for expected sub-array indexing
        Edm2WFEpad = pad_crop(Edm2WFE, mp.dm1.compact.NdmPad)  # Pad or crop for expected sub-array indexing
    
        # Propagate each actuator from DM1 through the optical system
        Gindex = 0  # initialize index counter
        for iact in mp.dm1.act_ele:
            # Compute only for influence functions that are not zeroed out
            if np.sum(np.abs(mp.dm1.compact.inf_datacube[:, :, iact])) > 1e-12:
                
                # x- and y- coordinate indices of the padded influence function in the full padded pupil
                x_box_AS_ind = np.arange(mp.dm1.compact.xy_box_lowerLeft_AS[0,iact], mp.dm1.compact.xy_box_lowerLeft_AS[0, iact]+NboxPad1AS, dtype=np.int) # x-indices in pupil arrays for the box
                y_box_AS_ind = np.arange(mp.dm1.compact.xy_box_lowerLeft_AS[1,iact], mp.dm1.compact.xy_box_lowerLeft_AS[1, iact]+NboxPad1AS, dtype=np.int) # y-indices in pupil arrays for the box
                indBoxAS = np.ix_(y_box_AS_ind, x_box_AS_ind)
                # x- and y- coordinates of the UN-padded influence function in the full padded pupil
                x_box = mp.dm1.compact.x_pupPad[x_box_AS_ind] # full pupil x-coordinates of the box 
                y_box = mp.dm1.compact.y_pupPad[y_box_AS_ind] # full pupil y-coordinates of the box
                
                # Propagate from DM1 to DM2, and then back to P2
                dEbox = (mirrorFac*2*np.pi*1j/wvl)*pad_crop((mp.dm1.VtoH.reshape(mp.dm1.Nact**2)[iact])*np.squeeze(mp.dm1.compact.inf_datacube[:,:,iact]),NboxPad1AS) # Pad influence function at DM1 for angular spectrum propagation.
                dEbox = fp.ptp(dEbox*Edm1pad[np.ix_(y_box_AS_ind,x_box_AS_ind)], mp.P2.compact.dx*NboxPad1AS,wvl, mp.d_dm1_dm2) # forward propagate to DM2 and apply DM2 E-field
                dEP2box = fp.ptp(dEbox*Edm2WFEpad[np.ix_(y_box_AS_ind,x_box_AS_ind)]*DM2stop[np.ix_(y_box_AS_ind,x_box_AS_ind)]*np.exp(mirrorFac*2*np.pi*1j/wvl*DM2surf[np.ix_(y_box_AS_ind,x_box_AS_ind)]), mp.P2.compact.dx*NboxPad1AS,wvl,-1*(mp.d_dm1_dm2 + mp.d_P2_dm1) ) # back-propagate to DM1
#                dEbox = fp.ptp_inf_func(dEbox*Edm1pad[np.ix_(y_box_AS_ind,x_box_AS_ind)], mp.P2.compact.dx*NboxPad1AS,wvl, mp.d_dm1_dm2, mp.dm1.dm_spacing, mp.propMethodPTP) # forward propagate to DM2 and apply DM2 E-field
#                dEP2box = fp.ptp_inf_func(dEbox.*Edm2WFEpad[np.ix_(y_box_AS_ind,x_box_AS_ind)]*DM2stop(y_box_AS_ind,x_box_AS_ind).*exp(mirrorFac*2*np.pi*1j/wvl*DM2surf(y_box_AS_ind,x_box_AS_ind)), mp.P2.compact.dx*NboxPad1AS,wvl,-1*(mp.d_dm1_dm2 + mp.d_P2_dm1), mp.dm1.dm_spacing, mp.propMethodPTP ) # back-propagate to DM1
#
                # To simulate going forward to the next pupil plane (with the apodizer) most efficiently, 
                # First, back-propagate the apodizer (by rotating 180-degrees) to the previous pupil.
                # Second, negate the coordinates of the box used.
                dEP2box = apodReimaged[indBoxAS]*dEP2box # Apply 180deg-rotated SP mask.
                dEP3box = np.rot90(dEP2box, k=2*mp.Nrelay2to3) # Forward propagate the cropped box by rotating 180 degrees mp.Nrelay2to3 times.
                # Negate and reverse coordinate values to effectively rotate by 180 degrees. No change if 360 degree rotation.
                if np.mod(mp.Nrelay2to3, 2) == 1:
                    x_box = -1*x_box[::-1]
                    y_box = -1*y_box[::-1]
               
                # Matrices for the MFT from the pupil P3 to the focal plane mask
                rect_mat_pre = (np.exp(-2*np.pi*1j*np.outer(mp.F3.compact.etas,y_box)/(wvl*mp.fl)))*np.sqrt(mp.P2.compact.dx*mp.P2.compact.dx)*np.sqrt(mp.F3.compact.dxi*mp.F3.compact.deta)/(wvl*mp.fl)
                rect_mat_post = (np.exp(-2*np.pi*1j*np.outer(x_box, mp.F3.compact.xis)/(wvl*mp.fl)))
                
                EF3inc = rect_mat_pre @ dEP3box @ rect_mat_post  # MFT to FPM
                
                if mp.coro.upper() in ('LC', 'APLC', 'HLC'):
                    # Propagate through (1 - FPM) for Babinet's principle
                    EF3 = (transOuterFPM-fpm) * EF3inc
        
                    # MFT to LS ("Sub" name for Subtrahend part of the Lyot-plane E-field)
                    EP4sub = fp.mft_f2p(EF3, mp.fl, wvl, mp.F3.compact.dxi, mp.F3.compact.deta, mp.P4.compact.dx, mp.P4.compact.Narr, mp.centering)
                    EP4sub = fp.relay(EP4sub, mp.Nrelay3to4-1, mp.centering)
                    
                    # Full Lyot plane pupil (for Babinet)
                    EP4noFPM = np.zeros((mp.dm1.compact.NdmPad, mp.dm1.compact.NdmPad),dtype=np.complex)
                    EP4noFPM[indBoxAS] = dEP2box # Propagating the E-field from P2 to P4 without masks gives the same E-field. 
                    EP4noFPM = fp.relay(EP4noFPM, mp.Nrelay2to3+mp.Nrelay3to4, mp.centering) # Get the correct orientation 
                    EP4noFPM = pad_crop(EP4noFPM, mp.P4.compact.Narr)  # Crop down to the size of the Lyot stop opening
                    EP4 = transOuterFPM*EP4noFPM - EP4sub  # Babinet's principle to get E-field at Lyot plane
                
                elif mp.coro.upper() in ('FLC', 'SPLC'):
                    EF3 = fpm * EF3inc  # Apply FPM
                    
                    # MFT to Lyot plane
                    EP4 = fp.mft_f2p(EF3, mp.fl,wvl, mp.F3.compact.dxi, mp.F3.compact.deta, mp.P4.compact.dx, mp.P4.compact.Narr, mp.centering)
                    EP4 = fp.relay(EP4, mp.Nrelay3to4-1, mp.centering)  # Get the correct orientation
                    
                EP4 *= mp.P4.compact.croppedMask  # Apply Lyot stop
    
                # MFT to camera
                EP4 = fp.relay(EP4, mp.NrelayFend, mp.centering) # Rotate the final image 180 degrees if necessary
                EFend = fp.mft_p2f(EP4, mp.fl,wvl, mp.P4.compact.dx, mp.Fend.dxi, mp.Fend.Nxi, mp.Fend.deta, mp.Fend.Neta, mp.centering)
                
                Gzdl[:, Gindex] = EFend[mp.Fend.corr.maskBool]/np.sqrt(mp.Fend.compact.I00[modvar.sbpIndex])

            Gindex += 1

    """ ---------- DM2 ---------- """
    if idm == 2:
        Gzdl = np.zeros((mp.Fend.corr.Npix, mp.dm2.Nele), dtype=np.complex)
        
        # Two array sizes (at same resolution) of influence functions for MFT and angular spectrum
        NboxPad2AS = int(mp.dm2.compact.NboxAS)
        mp.dm2.compact.xy_box_lowerLeft_AS = mp.dm2.compact.xy_box_lowerLeft - (NboxPad2AS-mp.dm2.compact.Nbox)/2 # Account for the padding of the influence function boxes
        
        apodReimaged = pad_crop(apodReimaged, mp.dm2.compact.NdmPad)
        DM2stopPad = pad_crop(DM2stop, mp.dm2.compact.NdmPad)
        Edm2WFEpad = pad_crop(Edm2WFE, mp.dm2.compact.NdmPad)
    
        # Propagate full field to DM2 before back-propagating in small boxes
        Edm2inc = pad_crop(fp.ptp(Edm1out, mp.compact.NdmPad*mp.P2.compact.dx, wvl, mp.d_dm1_dm2), mp.dm2.compact.NdmPad)  # E-field incident upon DM2
        Edm2inc = pad_crop(Edm2inc, mp.dm2.compact.NdmPad)
        Edm2 = DM2stopPad*Edm2WFEpad*Edm2inc*np.exp(mirrorFac*2*np.pi*1j/wvl*pad_crop(DM2surf, mp.dm2.compact.NdmPad))  # Initial E-field at DM2 including its own phase contribution
        
        # Propagate each actuator from DM2 through the rest of the optical system
        Gindex = 0  # initialize index counter
        for iact in mp.dm2.act_ele:
            if np.sum(np.abs(mp.dm2.compact.inf_datacube[:, :, iact])) > 1e-12:  # Only compute for acutators specified for use or for influence functions that are not zeroed out
    
                # x- and y- coordinates of the padded influence function in the full padded pupil
                x_box_AS_ind = np.arange(mp.dm2.compact.xy_box_lowerLeft_AS[0, iact], mp.dm2.compact.xy_box_lowerLeft_AS[0, iact]+NboxPad2AS, dtype=np.int) # x-indices in pupil arrays for the box
                y_box_AS_ind = np.arange(mp.dm2.compact.xy_box_lowerLeft_AS[1, iact], mp.dm2.compact.xy_box_lowerLeft_AS[1, iact]+NboxPad2AS, dtype=np.int) # y-indices in pupil arrays for the box
                indBoxAS = np.ix_(y_box_AS_ind, x_box_AS_ind)
                # x- and y- coordinates of the UN-padded influence function in the full padded pupil
                x_box = mp.dm2.compact.x_pupPad[x_box_AS_ind]  # full pupil x-coordinates of the box 
                y_box = mp.dm2.compact.y_pupPad[y_box_AS_ind]  # full pupil y-coordinates of the box 
                
                dEbox = (mp.dm2.VtoH.reshape(mp.dm2.Nact**2)[iact])*(mirrorFac*2*np.pi*1j/wvl)*pad_crop(np.squeeze(mp.dm2.compact.inf_datacube[:, :, iact]), NboxPad2AS) # the padded influence function at DM2
                dEP2box = fp.ptp(dEbox*Edm2[indBoxAS], mp.P2.compact.dx*NboxPad2AS, wvl, -1*(mp.d_dm1_dm2 + mp.d_P2_dm1)) # back-propagate to pupil P2
#                dEP2box = ptp_inf_func(dEbox.*Edm2(y_box_AS_ind,x_box_AS_ind), mp.P2.compact.dx*NboxPad2AS,wvl,-1*(mp.d_dm1_dm2 + mp.d_P2_dm1), mp.dm2.dm_spacing, mp.propMethodPTP); # back-propagate to pupil P2
                
                # To simulate going forward to the next pupil plane (with the apodizer) most efficiently, 
                # First, back-propagate the apodizer (by rotating 180-degrees) to the previous pupil.
                # Second, negate the coordinates of the box used.
                dEP2box = apodReimaged[indBoxAS]*dEP2box  # Apply 180deg-rotated SP mask.
                dEP3box = np.rot90(dEP2box, k=2*mp.Nrelay2to3)  # Forward propagate the cropped box by rotating 180 degrees mp.Nrelay2to3 times.
                # Negate and rotate coordinates to effectively rotate by 180 degrees. No change if 360 degree rotation.
                if np.mod(mp.Nrelay2to3, 2) == 1:
                    x_box = -1*x_box[::-1]
                    y_box = -1*y_box[::-1]
       
                # Matrices for the MFT from the pupil P3 to the focal plane mask
                rect_mat_pre = np.exp(-2*np.pi*1j*np.outer(mp.F3.compact.etas, y_box)/(wvl*mp.fl))*np.sqrt(mp.P2.compact.dx*mp.P2.compact.dx)*np.sqrt(mp.F3.compact.dxi*mp.F3.compact.deta)/(wvl*mp.fl)
                rect_mat_post = np.exp(-2*np.pi*1j*np.outer(x_box, mp.F3.compact.xis)/(wvl*mp.fl))
    
                EF3inc = rect_mat_pre @ dEP3box @ rect_mat_post  # MFT to FPM
                
                if mp.coro.upper() in ('LC', 'APLC', 'HLC'):
                    # Propagate through (1 - fpm) for Babinet's principle
                    EF3 = (transOuterFPM-fpm) * EF3inc
        
                    # MFT to LS ("Sub" name for Subtrahend part of the Lyot-plane E-field)
                    EP4sub = fp.mft_f2p(EF3, mp.fl, wvl, mp.F3.compact.dxi, mp.F3.compact.deta, mp.P4.compact.dx, mp.P4.compact.Narr, mp.centering) # Subtrahend term for the Lyot plane E-field    
                    EP4sub = fp.relay(EP4sub, mp.Nrelay3to4-1, mp.centering) # Get the correct orientation
                                    
                    EP4noFPM = np.zeros((mp.dm2.compact.NdmPad, mp.dm2.compact.NdmPad), dtype=np.complex)
                    EP4noFPM[indBoxAS] = dEP2box  # Propagating the E-field from P2 to P4 without masks gives the same E-field.
                    EP4noFPM = fp.relay(EP4noFPM, mp.Nrelay2to3+mp.Nrelay3to4, mp.centering) # Get the number or re-imaging relays between pupils P3 and P4. 
                    EP4noFPM = pad_crop(EP4noFPM, mp.P4.compact.Narr)  # Crop down to the size of the Lyot stop opening
                    EP4 = transOuterFPM*EP4noFPM - EP4sub  # Babinet's principle to get E-field at Lyot plane
                
                elif mp.coro.upper() in ('FLC', 'SPLC'):

                    EF3 = fpm * EF3inc  # Apply FPM
        
                    # MFT to LS ("Sub" name for Subtrahend part of the Lyot-plane E-field)
                    EP4 = fp.mft_f2p(EF3, mp.fl, wvl, mp.F3.compact.dxi, mp.F3.compact.deta, mp.P4.compact.dx, mp.P4.compact.Narr, mp.centering)
                    EP4 = fp.relay(EP4, mp.Nrelay3to4-1, mp.centering)
                
                EP4 *= mp.P4.compact.croppedMask  # Apply Lyot stop
    
                # MFT to detector
                EP4 = fp.relay(EP4, mp.NrelayFend, mp.centering)  # Rotate the final image 180 degrees if necessary
                EFend = fp.mft_p2f(EP4, mp.fl, wvl, mp.P4.compact.dx, mp.Fend.dxi, mp.Fend.Nxi, mp.Fend.deta, mp.Fend.Neta, mp.centering)
                
                Gzdl[:, Gindex] = EFend[mp.Fend.corr.maskBool]/np.sqrt(mp.Fend.compact.I00[modvar.sbpIndex])

            Gindex += 1

    """ ---------- DM9 (HLC only) ---------- """
    if idm == 9:
        Gzdl = np.zeros((mp.Fend.corr.Npix, mp.dm9.Nele), dtype=complex)
        Nbox9 = int(mp.dm9.compact.Nbox)

        # Adjust the step size in the Jacobian, then divide back out. Used for
        # helping counteract effect of discretization.
        if not hasattr(mp.dm9, 'stepFac'):
            stepFac = 20
        else:
            stepFac = mp.dm9.stepFac
        
        # Propagate from DM1 to DM2, and apply DM2 surface and aperture stop
        Edm2 = Edm2WFE * DM2stop * np.exp(mirrorFac*2*np.pi*1j*DM2surf/wvl) * \
            fp.ptp(Edm1out, mp.P2.compact.dx*NdmPad, wvl, mp.d_dm1_dm2)
        
        # Back-propagate to pupil P2
        dz2 = mp.d_P2_dm1 + mp.d_dm1_dm2
        if dz2 < 10*wvl:
            EP2eff = Edm2
        else:
            EP2eff = fp.ptp(Edm2, mp.P2.compact.dx*NdmPad, wvl, -dz2)
        
        # Rotate 180 degrees mp.Nrelay2to3 times to go from pupil P2 to P3
        EP3 = fp.relay(EP2eff, mp.Nrelay2to3, mp.centering)
    
        # Apply apodizer mask
        if mp.flagApod:
            EP3 = mp.P3.compact.mask * pad_crop(EP3, mp.P1.compact.Narr)
        
        # MFT from pupil P3 to FPM (at focus F3)
        EF3inc = fp.mft_p2f(EP3, mp.fl, wvl, mp.P2.compact.dx, mp.F3.compact.dxi, mp.F3.compact.Nxi, mp.F3.compact.deta, mp.F3.compact.Neta, mp.centering)
        EF3inc = pad_crop(EF3inc, mp.dm9.compact.NdmPad)
        # Coordinates for metal thickness and dielectric thickness
        DM8transIndAll = falco.hlc.discretize_fpm_surf(mp.dm8.surf, mp.t_metal_nm_vec, mp.dt_metal_nm)  # All of the mask
    
        # Propagate each actuator from DM2 through the rest of the optical system
        Gindex = 0  # initialize index counter
        for iact in mp.dm9.act_ele:
            if np.sum(np.abs(mp.dm9.compact.inf_datacube[:, :, iact])) > 1e-12:  # Only compute for acutators specified for use or for influence functions that are not zeroed out
    
                # xi- and eta- coordinates in the full FPM portion of the focal plane
                xyLL = mp.dm9.compact.xy_box_lowerLeft[:, iact]
                xi_box_ind = np.arange(xyLL[0], xyLL[0]+Nbox9, dtype=int)  # xi-indices in focal arrays for the box
                eta_box_ind = np.arange(xyLL[1], xyLL[1]+Nbox9, dtype=int)  # eta-indices in focal arrays for the box
                indBox = np.ix_(eta_box_ind, xi_box_ind)
                xi_box = mp.dm9.compact.x_pupPad[xi_box_ind]
                eta_box = mp.dm9.compact.y_pupPad[eta_box_ind]
                
                # Obtain values for the "poked" FPM's complex transmission (only in the sub-array where poked)
                Nxi = Nbox9
                Neta = Nbox9
                DM9surfCropNew = stepFac*mp.dm9.VtoH[iact]*mp.dm9.compact.inf_datacube[:, :, iact] + mp.dm9.surf[indBox]  # New DM9 surface profile in the poked region (meters)
                DM9transInd = falco.hlc.discretize_fpm_surf(DM9surfCropNew, mp.t_diel_nm_vec,  mp.dt_diel_nm)
                DM8transInd = DM8transIndAll[indBox]  # Cropped region of the FPM.
    
                # Look up table to compute complex transmission coefficient of the FPM at each pixel
                fpmPoked = np.zeros((Neta, Nxi), dtype=complex)  # Initialize output array of FPM's complex transmission
                for ix in range(Nxi):
                    for iy in range(Neta):
                        ind_metal = DM8transInd[iy, ix]
                        ind_diel = DM9transInd[iy, ix]
                        fpmPoked[iy, ix] = mp.complexTransCompact[ind_diel, ind_metal, modvar.sbpIndex]

    
                dEF3box = ((transOuterFPM-fpmPoked) - (transOuterFPM-fpm[indBox])) * EF3inc[indBox]  # Delta field (in a small region) at the FPM
    
                # Matrices for the MFT from the FPM stamp to the Lyot stop
                rect_mat_pre = np.exp(-2*np.pi*1j*np.outer(mp.P4.compact.ys, eta_box)/(wvl*mp.fl)) *\
                    np.sqrt(mp.P4.compact.dx*mp.P4.compact.dx)*np.sqrt(mp.F3.compact.dxi*mp.F3.compact.deta)/(wvl*mp.fl)
                rect_mat_post = np.exp(-2*np.pi*1j*np.outer(xi_box, mp.P4.compact.xs)/(wvl*mp.fl))
    
                # MFT from FPM to Lyot stop (Nominal term transOuterFPM*EP4noFPM subtracts out to 0 since it ignores the FPM change).
                EP4 = 0 - rect_mat_pre @ dEF3box @ rect_mat_post  # MFT from FPM (F3) to Lyot stop plane (P4)
                EP4 = fp.relay(EP4, mp.Nrelay3to4-1, mp.centering)
                EP4 = mp.P4.compact.croppedMask * EP4  # Apply Lyot stop
    
                # MFT to final focal plane
                EP4 = fp.relay(EP4, mp.NrelayFend, mp.centering)
                EFend = fp.mft_p2f(EP4, mp.fl, wvl, mp.P4.compact.dx, mp.Fend.dxi, mp.Fend.Nxi, mp.Fend.deta, mp.Fend.Neta, mp.centering)
    
                Gzdl[:, Gindex] = mp.dm9.act_sens / stepFac * mp.dm9.weight*EFend[mp.Fend.corr.maskBool] / np.sqrt(mp.Fend.compact.I00[modvar.sbpIndex])

            Gindex += 1
        
    return Gzdl

    
def vortex(mp, im, idm):
    """
    Differential model used to compute ctrl Jacobian for vortex coronagraph.

    Specialized compact model used to compute the DM response matrix, aka the
    control Jacobian for a vortex coronagraph. Can include an apodizer, making
    it an apodized vortex coronagraph (AVC). Does not include unknown
    aberrations of the full, "truth" model. This model propagates the
    first-order Taylor expansion of the phase from the poke of each actuator
    of the deformable mirror.

    Parameters
    ----------
    mp : ModelParameters
        Structure containing optical model parameters

    Returns
    -------
    Gzdl : numpy ndarray
        Complex-valued, 2-D array containing the Jacobian for the
        specified Zernike mode, DM number, and wavelength.

    """
    modvar = falco.config.Object()  # Initialize the new structure
    modvar.sbpIndex = mp.jac.sbp_inds[im]
    modvar.zernIndex = mp.jac.zern_inds[im]
    
    wvl = mp.sbp_centers[modvar.sbpIndex]
    mirrorFac = 2.  # Phase change is twice the DM surface height.
    NdmPad = int(mp.compact.NdmPad)

    # Minimum FPM resolution for Jacobian calculations (in pixels per lambda/D)
    minPadFacVortex = 8

    # Get FPM charge
    if type(mp.F3.VortexCharge) == np.ndarray:
        # Passing an array for mp.F3.VortexCharge with
        # corresponding wavelengths mp.F3.VortexCharge_lambdas
        # represents a chromatic vortex FPM
        if mp.F3.VortexCharge.size == 1:
            charge = mp.F3.VortexCharge
        else:
            np.interp(wvl, mp.F3.VortexCharge_lambdas, mp.F3.VortexCharge,
                      'linear', 'extrap')
    elif type(mp.F3.VortexCharge) == int or type(mp.F3.VortexCharge) == float:
        # single value indicates fully achromatic mask
        charge = mp.F3.VortexCharge
    else:
        raise TypeError("mp.F3.VortexCharge must be an int, float, or numpy ndarray.")

    """Input E-fields"""
    Ein = np.squeeze(mp.P1.compact.E[:, :, modvar.sbpIndex])
    
    # Apply a Zernike (in amplitude) at input pupil
    # Used only for Zernike sensitivity control, which requires the perfect
    # E-field of the differential Zernike term.
    if not modvar.zernIndex == 1:
        indsZnoll = modvar.zernIndex  # Just send in 1 Zernike mode
        zernMat = np.squeeze(falco.zern.gen_norm_zern_maps(mp.P1.compact.Nbeam,
                                                    mp.centering, indsZnoll))
        zernMat = pad_crop(zernMat, mp.P1.compact.Narr)
        Ein = Ein*zernMat*(2*np.pi/wvl) * \
            mp.jac.Zcoef[mp.jac.zerns == modvar.zernIndex]

    """ Masks and DM surfaces """
    pupil = pad_crop(mp.P1.compact.mask, NdmPad)
    Ein = pad_crop(Ein, NdmPad)

    # Re-image the apodizer from pupil P3 back to pupil P2.
    if(mp.flagApod):
        apodReimaged = pad_crop(mp.P3.compact.mask, NdmPad)
        apodReimaged = fp.relay(apodReimaged, mp.Nrelay2to3, mp.centering)
    else:
        apodReimaged = np.ones((NdmPad, NdmPad))

    # Compute the DM surfaces for the current DM commands
    if any(mp.dm_ind == 1):
        DM1surf = pad_crop(mp.dm1.compact.surfM, NdmPad)
        # DM1surf = falco.dm.gen_surf_from_act(mp.dm1, mp.dm1.compact.dx, NdmPad)
    else:
        DM1surf = np.zeros((NdmPad, NdmPad))
    if any(mp.dm_ind == 2):
        DM2surf = pad_crop(mp.dm2.compact.surfM, NdmPad)
        # DM2surf = falco.dm.gen_surf_from_act(mp.dm2, mp.dm2.compact.dx, NdmPad)
    else:
        DM2surf = np.zeros((NdmPad, NdmPad))

    if(mp.flagDM1stop):
        DM1stop = pad_crop(mp.dm1.compact.mask, NdmPad)
    else:
        DM1stop = np.ones((NdmPad, NdmPad))
    if(mp.flagDM2stop):
        DM2stop = pad_crop(mp.dm2.compact.mask, NdmPad)
    else:
        DM2stop = np.ones((NdmPad, NdmPad))

    # This block is for BMC surface error testing
    if(mp.flagDMwfe):
        if any(mp.dm_ind == 1):
            Edm1WFE = np.exp(2*np.pi*1j/wvl*pad_crop(mp.dm1.compact.wfe,
                                                     NdmPad, 'extrapval', 0))
        else:
            Edm1WFE = np.ones((NdmPad, NdmPad))
        if any(mp.dm_ind == 2):
            Edm2WFE = np.exp(2*np.pi*1j/wvl*pad_crop(mp.dm2.compact.wfe,
                                                     NdmPad, 'extrapval', 0))
        else:
            Edm2WFE = np.ones((NdmPad, NdmPad))
    else:
        Edm1WFE = np.ones((NdmPad, NdmPad))
        Edm2WFE = np.ones((NdmPad, NdmPad))
        
    """Propagation"""
    # Define pupil P1 and Propagate to pupil P2
    EP1 = pupil*Ein  # E-field at pupil plane P1
    EP2 = fp.relay(EP1, mp.Nrelay1to2, mp.centering)

    # Propagate from P2 to DM1, and apply DM1 surface and aperture stop
    if not (abs(mp.d_P2_dm1) == 0):  # E-field arriving at DM1
        Edm1 = fp.ptp(EP2, mp.P2.compact.dx*NdmPad, wvl, mp.d_P2_dm1)
    else:
        Edm1 = EP2
    Edm1out = Edm1*Edm1WFE*DM1stop*np.exp(mirrorFac*2*np.pi*1j*DM1surf/wvl)

    """ ---------- DM1 ---------- """
    if idm == 1:
        Gzdl = np.zeros((mp.Fend.corr.Npix, mp.dm1.Nele), dtype=np.complex)
        
        # Array size for planes P3, F3, and P4
        Nfft1 = int(2**falco.util.nextpow2(np.max(np.array([mp.dm1.compact.NdmPad, minPadFacVortex*mp.dm1.compact.Nbox])))) # Don't crop--but do pad if necessary.
        
        # Generate vortex FPM with fftshift already applied
        fftshiftVortex = fftshift(falco.mask.falco_gen_vortex_mask(charge, Nfft1))
    
        # Two array sizes (at same resolution) of influence functions for MFT and angular spectrum
        NboxPad1AS = int(mp.dm1.compact.NboxAS)  # array size for FFT-AS propagations from DM1->DM2->DM1
        mp.dm1.compact.xy_box_lowerLeft_AS = mp.dm1.compact.xy_box_lowerLeft - (mp.dm1.compact.NboxAS-mp.dm1.compact.Nbox)/2. # Adjust the sub-array location of the influence function for the added zero padding
    
        if any(mp.dm_ind == 2):
            DM2surf = pad_crop(DM2surf, mp.dm1.compact.NdmPad)
        else:
            DM2surf = np.zeros((mp.dm1.compact.NdmPad, mp.dm1.compact.NdmPad))
        if(mp.flagDM2stop):
            DM2stop = pad_crop(DM2stop, mp.dm1.compact.NdmPad)
        else:
            DM2stop = np.ones((mp.dm1.compact.NdmPad, mp.dm1.compact.NdmPad))
        apodReimaged = pad_crop(apodReimaged, mp.dm1.compact.NdmPad)
    
        Edm1pad = pad_crop(Edm1out, mp.dm1.compact.NdmPad)  # Pad or crop for expected sub-array indexing
        Edm2WFEpad = pad_crop(Edm2WFE, mp.dm1.compact.NdmPad)  # Pad or crop for expected sub-array indexing
    
        # Propagate each actuator from DM1 through the optical system
        Gindex = 0  # initialize index counter
        for iact in mp.dm1.act_ele:
            # Compute only for influence functions that are not zeroed out
            if np.sum(np.abs(mp.dm1.compact.inf_datacube[:, :, iact])) > 1e-12:
                
                # x- and y- coordinate indices of the padded influence function in the full padded pupil
                x_box_AS_ind = np.arange(mp.dm1.compact.xy_box_lowerLeft_AS[0, iact], mp.dm1.compact.xy_box_lowerLeft_AS[0, iact]+NboxPad1AS, dtype=np.int)  # x-indices in pupil arrays for the box
                y_box_AS_ind = np.arange(mp.dm1.compact.xy_box_lowerLeft_AS[1, iact], mp.dm1.compact.xy_box_lowerLeft_AS[1 ,iact]+NboxPad1AS, dtype=np.int)  # y-indices in pupil arrays for the box
                indBoxAS = np.ix_(y_box_AS_ind, x_box_AS_ind)
                # x- and y- coordinates of the UN-padded influence function in the full padded pupil
                x_box = mp.dm1.compact.x_pupPad[x_box_AS_ind]  # full pupil x-coordinates of the box 
                y_box = mp.dm1.compact.y_pupPad[y_box_AS_ind]  # full pupil y-coordinates of the box
                
                # Propagate from DM1 to DM2, and then back to P2
                dEbox = (mirrorFac*2*np.pi*1j/wvl)*pad_crop((mp.dm1.VtoH.reshape(mp.dm1.Nact**2)[iact])*np.squeeze(mp.dm1.compact.inf_datacube[:, :, iact]), NboxPad1AS) # Pad influence function at DM1 for angular spectrum propagation.
                dEbox = fp.ptp(dEbox*Edm1pad[indBoxAS], mp.P2.compact.dx*NboxPad1AS,wvl, mp.d_dm1_dm2) # forward propagate to DM2 and apply DM2 E-field
                dEP2box = fp.ptp(dEbox*Edm2WFEpad[indBoxAS]*DM2stop[indBoxAS]*np.exp(mirrorFac*2*np.pi*1j/wvl*DM2surf[indBoxAS]), mp.P2.compact.dx*NboxPad1AS,wvl,-1*(mp.d_dm1_dm2 + mp.d_P2_dm1)) # back-propagate to DM1
#                dEbox = fp.ptp_inf_func(dEbox*Edm1pad[np.ix_(y_box_AS_ind,x_box_AS_ind)], mp.P2.compact.dx*NboxPad1AS,wvl, mp.d_dm1_dm2, mp.dm1.dm_spacing, mp.propMethodPTP) # forward propagate to DM2 and apply DM2 E-field
#                dEP2box = fp.ptp_inf_func(dEbox.*Edm2WFEpad[np.ix_(y_box_AS_ind,x_box_AS_ind)]*DM2stop(y_box_AS_ind,x_box_AS_ind).*exp(mirrorFac*2*np.pi*1j/wvl*DM2surf(y_box_AS_ind,x_box_AS_ind)), mp.P2.compact.dx*NboxPad1AS,wvl,-1*(mp.d_dm1_dm2 + mp.d_P2_dm1), mp.dm1.dm_spacing, mp.propMethodPTP ) # back-propagate to DM1
#
                # To simulate going forward to the next pupil plane (with the apodizer) most efficiently, 
                # First, back-propagate the apodizer (by rotating 180-degrees) to the previous pupil.
                # Second, negate the coordinates of the box used.
                dEP2boxEff = apodReimaged[indBoxAS]*dEP2box  # Apply 180deg-rotated apodizer mask.
                # dEP3box = np.rot90(dEP2box,k=2*mp.Nrelay2to3) # Forward propagate the cropped box by rotating 180 degrees mp.Nrelay2to3 times.
                # # Negate and reverse coordinate values to effectively rotate by 180 degrees. No change if 360 degree rotation.

                # Re-insert the window around the influence function back into the full beam array.
                EP2eff = np.zeros((mp.dm1.compact.NdmPad, mp.dm1.compact.NdmPad), dtype=complex)
                EP2eff[indBoxAS] = dEP2boxEff
                
                # Forward propagate from P2 (effective) to P3
                EP3 = fp.relay(EP2eff, mp.Nrelay2to3, mp.centering)
                
                # Pad pupil P3 for FFT
                EP3pad = pad_crop(EP3, Nfft1)
                
                # FFT from P3 to Fend.and apply vortex
                EF3 = fftshiftVortex*fft2(fftshift(EP3pad))/Nfft1
                
                # FFT from Vortex FPM to Lyot Plane
                EP4 = fftshift(fft2(EF3))/Nfft1
                EP4 = fp.relay(EP4, mp.Nrelay3to4-1, mp.centering)  # Add more re-imaging relays if necessary
                if(Nfft1 > mp.P4.compact.Narr):
                    EP4 = mp.P4.compact.croppedMask*pad_crop(EP4, mp.P4.compact.Narr)  # Crop EP4 and then apply Lyot stop 
                else:
                    EP4 = pad_crop(mp.P4.compact.croppedMask, Nfft1)*EP4  # Crop the Lyot stop and then apply it.
                    pass
                
                # MFT to camera
                EP4 = fp.relay(EP4, mp.NrelayFend, mp.centering)  # Rotate the final image 180 degrees if necessary
                EFend = fp.mft_p2f(EP4, mp.fl, wvl, mp.P4.compact.dx, mp.Fend.dxi, mp.Fend.Nxi, mp.Fend.deta, mp.Fend.Neta, mp.centering)
                
                Gzdl[:, Gindex] = EFend[mp.Fend.corr.maskBool]/np.sqrt(mp.Fend.compact.I00[modvar.sbpIndex])

            Gindex += 1

    """ ---------- DM2 ---------- """
    if idm == 2:
        Gzdl = np.zeros((mp.Fend.corr.Npix, mp.dm2.Nele), dtype=np.complex)
        
        # Array size for planes P3, F3, and P4
        Nfft2 = int(2**falco.util.nextpow2(np.max(np.array([mp.dm2.compact.NdmPad, minPadFacVortex*mp.dm2.compact.Nbox])))) # Don't crop--but do pad if necessary.
        
        # Generate vortex FPM with fftshift already applied
        fftshiftVortex = fftshift(falco.mask.falco_gen_vortex_mask(charge, Nfft2))
    
        # Two array sizes (at same resolution) of influence functions for MFT and angular spectrum
        NboxPad2AS = int(mp.dm2.compact.NboxAS)
        mp.dm2.compact.xy_box_lowerLeft_AS = mp.dm2.compact.xy_box_lowerLeft - (NboxPad2AS-mp.dm2.compact.Nbox)/2 # Account for the padding of the influence function boxes
        
        apodReimaged = pad_crop(apodReimaged, mp.dm2.compact.NdmPad)
        DM2stopPad = pad_crop(DM2stop, mp.dm2.compact.NdmPad)
        Edm2WFEpad = pad_crop(Edm2WFE, mp.dm2.compact.NdmPad)
    
        # Propagate full field to DM2 before back-propagating in small boxes
        Edm2inc = pad_crop(fp.ptp(Edm1out, mp.compact.NdmPad*mp.P2.compact.dx,wvl, mp.d_dm1_dm2), mp.dm2.compact.NdmPad) # E-field incident upon DM2
        Edm2inc = pad_crop(Edm2inc, mp.dm2.compact.NdmPad);
        Edm2 = DM2stopPad * Edm2WFEpad * Edm2inc * np.exp(mirrorFac*2*np.pi*1j/wvl * pad_crop(DM2surf, mp.dm2.compact.NdmPad)) # Initial E-field at DM2 including its own phase contribution
        
        # Propagate each actuator from DM2 through the rest of the optical system
        Gindex = 0  # initialize index counter
        for iact in mp.dm2.act_ele:
            # Only compute for acutators specified for use or for influence functions that are not zeroed out
            if np.sum(np.abs(mp.dm2.compact.inf_datacube[:, :, iact])) > 1e-12:
    
                # x- and y- coordinates of the padded influence function in the full padded pupil
                x_box_AS_ind = np.arange(mp.dm2.compact.xy_box_lowerLeft_AS[0, iact], mp.dm2.compact.xy_box_lowerLeft_AS[0, iact]+NboxPad2AS, dtype=np.int)  # x-indices in pupil arrays for the box
                y_box_AS_ind = np.arange(mp.dm2.compact.xy_box_lowerLeft_AS[1, iact], mp.dm2.compact.xy_box_lowerLeft_AS[1, iact]+NboxPad2AS, dtype=np.int)  # y-indices in pupil arrays for the box
                indBoxAS = np.ix_(y_box_AS_ind, x_box_AS_ind)
#               # x- and y- coordinates of the UN-padded influence function in the full padded pupil
#                x_box = mp.dm2.compact.x_pupPad[x_box_AS_ind] # full pupil x-coordinates of the box
#                y_box = mp.dm2.compact.y_pupPad[y_box_AS_ind] # full pupil y-coordinates of the box
                
                dEbox = (mp.dm2.VtoH.reshape(mp.dm2.Nact**2)[iact])*(mirrorFac*2*np.pi*1j/wvl)*pad_crop(np.squeeze(mp.dm2.compact.inf_datacube[:, :, iact]), NboxPad2AS) # the padded influence function at DM2
                dEP2box = fp.ptp(dEbox*Edm2[indBoxAS], mp.P2.compact.dx*NboxPad2AS, wvl, -1*(mp.d_dm1_dm2 + mp.d_P2_dm1)) # back-propagate to pupil P2
#                dEP2box = ptp_inf_func(dEbox.*Edm2(y_box_AS_ind,x_box_AS_ind), mp.P2.compact.dx*NboxPad2AS,wvl,-1*(mp.d_dm1_dm2 + mp.d_P2_dm1), mp.dm2.dm_spacing, mp.propMethodPTP); # back-propagate to pupil P2
                
                # To simulate going forward to the next pupil plane (with the apodizer) most efficiently,
                # First, back-propagate the apodizer (by rotating 180-degrees) to the previous pupil.
                # Second, negate the coordinates of the box used.
                dEP2boxEff = apodReimaged[indBoxAS]*dEP2box
#                dEP3box = np.rot90(dEP2box,k=2*mp.Nrelay2to3) # Forward propagate the cropped box by rotating 180 degrees mp.Nrelay2to3 times.
#                # Negate and rotate coordinates to effectively rotate by 180 degrees. No change if 360 degree rotation.
#                if np.mod(mp.Nrelay2to3,2)==1: 
#                    x_box = -1*x_box[::-1]
#                    y_box = -1*y_box[::-1]
                
                EP2eff = np.zeros((mp.dm2.compact.NdmPad, mp.dm2.compact.NdmPad), dtype=complex)
                EP2eff[indBoxAS] = dEP2boxEff

                # Forward propagate from P2 (effective) to P3
                EP3 = fp.relay(EP2eff, mp.Nrelay2to3, mp.centering)
    
                # Pad pupil P3 for FFT
                EP3pad = pad_crop(EP3, Nfft2)
                
                # FFT from P3 to Fend.and apply vortex
                EF3 = fftshiftVortex*fft2(fftshift(EP3pad))/Nfft2
    
                # FFT from Vortex FPM to Lyot Plane
                EP4 = fftshift(fft2(EF3))/Nfft2
                EP4 = fp.relay(EP4, mp.Nrelay3to4-1, mp.centering)
                
                if(Nfft2 > mp.P4.compact.Narr):
                    EP4 = mp.P4.compact.croppedMask * pad_crop(EP4, mp.P4.compact.Narr)
                else:
                    EP4 = pad_crop(mp.P4.compact.croppedMask, Nfft2) * EP4

                # MFT to detector
                EP4 = fp.relay(EP4, mp.NrelayFend, mp.centering)
                EFend = fp.mft_p2f(EP4, mp.fl, wvl, mp.P4.compact.dx, mp.Fend.dxi, mp.Fend.Nxi, mp.Fend.deta, mp.Fend.Neta, mp.centering)
                
                Gzdl[:, Gindex] = EFend[mp.Fend.corr.maskBool] / \
                    np.sqrt(mp.Fend.compact.I00[modvar.sbpIndex])

            Gindex += 1
        
    return Gzdl



""" 
Wrapper for the MKL FFT routines. This implements very fast FFT on Intel
processors, much faster than the stock fftpack routines in numpy/scipy.

"""
from __future__ import division, print_function

import numpy as np
import ctypes as _ctypes
import os

from dftidefs import *

def load_libmkl():
    r"""Loads the MKL library if it can be found in the library load path.

    Raises
    ------
    ValueError
        If the MKL library cannot be found.

    """

    if os.name == 'posix':
        try:
            lib_mkl = os.getenv('LIBMKL')
            if lib_mkl is None:
                raise ValueError('LIBMKL environment variable not found')
            return _ctypes.cdll.LoadLibrary(lib_mkl)
        except:
            pass
        try:
            return _ctypes.cdll.LoadLibrary("libmkl_rt.dylib")
        except:
            raise ValueError('MKL Library not found')

    else:
        try:
            return _ctypes.cdll.LoadLibrary("mkl_rt.dll")
        except:
            raise ValueError('MKL Library not found')

mkl = load_libmkl()

def mkl_rfft(a, n=None, axis=-1, norm=None, direction='forward', out=None, scrambled=False):
    r"""Forward/backward 1D double-precision real-complex FFT.

    Uses the Intel MKL libraries distributed with Anaconda Python.
    Normalisation is different from Numpy!
    By default, allocates new memory like 'a' for output data.
    Returns the array containing output data.

    See Also
    --------
    rfft, irfft

    """

    if axis == -1:
        axis = a.ndim-1

    # This code only works for 1D and 2D arrays
    assert a.ndim < 3
    assert (axis < a.ndim and axis >= -1)
    assert (direction == 'forward' or direction == 'backward')

    # Convert input to complex data type if real (also memory copy)
    if direction == 'forward' and a.dtype != np.float32 and a.dtype != np.float64:
        if a.dtype == np.int64 or a.dtype == np.uint64:
            a = np.array(a, dtype=np.float64)
        else:
            a = np.array(a, dtype=np.float32)

    elif direction == 'backward' and a.dtype != np.complex128 and a.dtype != np.complex64:
        if a.dtype == np.int64 or a.dtype == np.uint64 or a.dtype == np.float64:
            a = np.array(a, dtype=np.complex128)
        else:
            a = np.array(a, dtype=np.complex64)


    order = 'C'
    if a.flags['F_CONTIGUOUS'] and not a.flags['C_CONTIGUOUS']:
        order = 'F'

    # Add zero padding or truncate if needed (incurs memory copy)
    if n is not None:
        m = n if direction == 'forward' else (n // 2 + 1)
        if a.shape[axis] < m:
            # pad axis with zeros
            pad_width = np.zeros((a.ndim, 2), dtype=np.int)
            pad_width[axis,1] = m - a.shape[axis]
            a = np.pad(a, pad_width, mode='constant')
        elif a.shape[axis] > m:
            # truncate along axis
            b = np.swapaxes(a, axis, 0)[:m,]
            a = np.swapaxes(b, 0, axis).copy()

    elif direction == 'forward':
        n = a.shape[axis]

    elif direction == 'backward':
        n = 2*(a.shape[axis]-1)


    # determine output type
    if direction == 'backward':
        out_type = np.float64
        if a.dtype == np.complex64:
            out_type = np.float32
    elif direction == 'forward':
        out_type = np.complex128
        if a.dtype == np.float32:
            out_type = np.complex64

    # Configure output array
    assert a is not out
    if out is not None:
        assert out.dtype == out_type
        for i in range(a.ndim):
            if i != axis:
                assert a.shape[i] == out.shape[i]
        if direction == 'forward':
            assert (n // 2 + 1) == out.shape[axis]
        else:
            assert out.shape[axis] == n
        assert not np.may_share_memory(a, out)
    else:
        size = list(a.shape)
        size[axis] = n // 2 + 1 if direction == 'forward' else n
        out = np.empty(size, dtype=out_type, order=order)

    # Define length, number of transforms strides
    length = _ctypes.c_int(n)
    n_transforms = _ctypes.c_int(np.prod(a.shape) // a.shape[axis])

    # For strides, the C type used *must* be long
    strides = (_ctypes.c_long*2)(0, a.strides[axis] // a.itemsize)
    if a.ndim == 2:
        if axis == 0:
            distance = _ctypes.c_int(a.strides[1] // a.itemsize)
            out_distance = _ctypes.c_int(out.strides[1] // out.itemsize)
        else:
            distance = _ctypes.c_int(a.strides[0] // a.itemsize)
            out_distance = _ctypes.c_int(out.strides[0] // out.itemsize)

    double_precision = True
    if (direction == 'forward' and a.dtype == np.float32) or (direction == 'backward' and a.dtype == np.complex64):
        double_precision = False

    # Create the description handle
    Desc_Handle = _ctypes.c_void_p(0)
    if not double_precision:
        mkl.DftiCreateDescriptor(_ctypes.byref(Desc_Handle), DFTI_SINGLE, DFTI_REAL, _ctypes.c_int(1), length)
    else:
        mkl.DftiCreateDescriptor(_ctypes.byref(Desc_Handle), DFTI_DOUBLE, DFTI_REAL, _ctypes.c_int(1), length)

    # set the storage type
    mkl.DftiSetValue(Desc_Handle, DFTI_CONJUGATE_EVEN_STORAGE, DFTI_COMPLEX_COMPLEX)

    # set normalization factor
    if norm == 'ortho':
        scale = _ctypes.c_double(1 / np.sqrt(n))
        mkl.DftiSetValue(Desc_Handle, DFTI_FORWARD_SCALE, scale)
        mkl.DftiSetValue(Desc_Handle, DFTI_BACKWARD_SCALE, scale)
    elif norm is None:
        scale = _ctypes.c_double(1. / n)
        s = mkl.DftiSetValue(Desc_Handle, DFTI_BACKWARD_SCALE, scale)

    # set all values if necessary
    if a.ndim != 1:
        mkl.DftiSetValue(Desc_Handle, DFTI_NUMBER_OF_TRANSFORMS, n_transforms)
        mkl.DftiSetValue(Desc_Handle, DFTI_INPUT_DISTANCE, distance)
        mkl.DftiSetValue(Desc_Handle, DFTI_OUTPUT_DISTANCE, out_distance)
        mkl.DftiSetValue(Desc_Handle, DFTI_INPUT_STRIDES, _ctypes.byref(strides))
        mkl.DftiSetValue(Desc_Handle, DFTI_OUTPUT_STRIDES, _ctypes.byref(strides))

    if scrambled:
        s = mkl.DftiSetValue(Desc_Handle, DFTI_ORDERING, DFTI_BACKWARD_SCRAMBLED)

    if direction == 'forward':
        fft_func = mkl.DftiComputeForward
    elif direction == 'backward':
        fft_func = mkl.DftiComputeBackward
    else:
        assert False

    # Not-in-place FFT
    mkl.DftiSetValue(Desc_Handle, DFTI_PLACEMENT, DFTI_NOT_INPLACE)

    mkl.DftiCommitDescriptor(Desc_Handle)
    fft_func(Desc_Handle, a.ctypes.data_as(_ctypes.c_void_p), out.ctypes.data_as(_ctypes.c_void_p) )

    mkl.DftiFreeDescriptor(_ctypes.byref(Desc_Handle))

    return out


def mkl_fft(a, n=None, axis=-1, norm=None, direction='forward', out=None, scrambled=False):
    r"""Forward/backward 1D single- or double-precision FFT.

    Uses the Intel MKL libraries distributed with Anaconda Python.
    Normalisation is different from Numpy!
    By default, allocates new memory like 'a' for output data.
    Returns the array containing output data.

    See Also
    --------
    fft, ifft

    """

    # This code only works for 1D and 2D arrays
    assert a.ndim < 3
    assert axis < a.ndim and axis >= -1

    # Add zero padding if needed (incurs memory copy)
    '''
    if n is not None and n != a.shape[axis]:
        pad_width = np.zeros((a.ndim, 2), dtype=np.int)
        pad_width[axis,1] = n - a.shape[axis]
        a = np.pad(a, pad_width, mode='constant')
    '''

    if n is not None:
        if a.shape[axis] < n:
            # pad axis with zeros
            pad_width = np.zeros((a.ndim, 2), dtype=np.int)
            pad_width[axis,1] = n - a.shape[axis]
            a = np.pad(a, pad_width, mode='constant')
        elif a.shape[axis] > n:
            # truncate along axis
            b = np.swapaxes(a, axis, -1)[...,:n]
            a = np.swapaxes(b, -1, axis).copy()

    # Convert input to complex data type if real (also memory copy)
    if a.dtype != np.complex128 and a.dtype != np.complex64:
        if a.dtype == np.int64 or a.dtype == np.uint64 or a.dtype == np.float64:
            a = np.array(a, dtype=np.complex128)
        else:
            a = np.array(a, dtype=np.complex64)

    # Configure in-place vs out-of-place
    inplace = False
    if out is a:
        inplace = True
    elif out is not None:
        assert out.dtype == a.dtype
        assert a.shape == out.shape
        assert not np.may_share_memory(a, out)
    else:
        out = np.empty_like(a)

    # Define length, number of transforms strides
    length = _ctypes.c_int(a.shape[axis])
    n_transforms = _ctypes.c_int(np.prod(a.shape) // a.shape[axis])
    
    # For strides, the C type used *must* be long
    strides = (_ctypes.c_long*2)(0, a.strides[axis] // a.itemsize)
    if a.ndim == 2:
        if axis == 0:
            distance = _ctypes.c_int(a.strides[1] // a.itemsize)
        else:
            distance = _ctypes.c_int(a.strides[0] // a.itemsize)

    # Create the description handle
    Desc_Handle = _ctypes.c_void_p(0)
    if a.dtype == np.complex64:
        mkl.DftiCreateDescriptor(_ctypes.byref(Desc_Handle), DFTI_SINGLE, DFTI_COMPLEX, _ctypes.c_int(1), length)
    elif a.dtype == np.complex128:
        mkl.DftiCreateDescriptor(_ctypes.byref(Desc_Handle), DFTI_DOUBLE, DFTI_COMPLEX, _ctypes.c_int(1), length)

    # Set normalization factor
    if norm == 'ortho':
        if a.dtype == np.complex64:
            scale = _ctypes.c_float(1 / np.sqrt(a.shape[axis]))
        else:
            scale = _ctypes.c_double(1 / np.sqrt(a.shape[axis]))
        mkl.DftiSetValue(Desc_Handle, DFTI_FORWARD_SCALE, scale)
        mkl.DftiSetValue(Desc_Handle, DFTI_BACKWARD_SCALE, scale)
    elif norm is None:
        if a.dtype == np.complex64:
            scale = _ctypes.c_float(1. / a.shape[axis])
        else:
            scale = _ctypes.c_double(1. / a.shape[axis])
        mkl.DftiSetValue(Desc_Handle, DFTI_BACKWARD_SCALE, scale)

    # set all values if necessary
    if a.ndim != 1:
        mkl.DftiSetValue(Desc_Handle, DFTI_NUMBER_OF_TRANSFORMS, n_transforms)
        mkl.DftiSetValue(Desc_Handle, DFTI_INPUT_DISTANCE, distance)
        mkl.DftiSetValue(Desc_Handle, DFTI_OUTPUT_DISTANCE, distance)
        mkl.DftiSetValue(Desc_Handle, DFTI_INPUT_STRIDES, _ctypes.byref(strides))
        mkl.DftiSetValue(Desc_Handle, DFTI_OUTPUT_STRIDES, _ctypes.byref(strides))

    if scrambled:
        s = mkl.DftiSetValue(Desc_Handle, DFTI_ORDERING, DFTI_BACKWARD_SCRAMBLED)
        DftiErrorMessage(s)

    if direction == 'forward':
        fft_func = mkl.DftiComputeForward
    elif direction == 'backward':
        fft_func = mkl.DftiComputeBackward
    else:
        assert False

    if inplace:
        # In-place FFT
        mkl.DftiCommitDescriptor(Desc_Handle)
        fft_func(Desc_Handle, a.ctypes.data_as(_ctypes.c_void_p) )

    else:
        # Not-in-place FFT
        mkl.DftiSetValue(Desc_Handle, DFTI_PLACEMENT, DFTI_NOT_INPLACE)

        mkl.DftiCommitDescriptor(Desc_Handle)
        fft_func(Desc_Handle, a.ctypes.data_as(_ctypes.c_void_p), out.ctypes.data_as(_ctypes.c_void_p) )


    mkl.DftiFreeDescriptor(_ctypes.byref(Desc_Handle))

    return out

def proper_fft2(a, norm=None, direction='forward', mkl_dir=None, fft_nthreads=0):
    r"""Forward/backward 2D single- or double-precision FFT.

    Uses the Intel MKL libraries distributed with Enthought Python.
    Normalisation is different from Numpy!
    By default, allocates new memory like 'a' for output data.
    Returns the array containing output data.

    See Also
    --------
    fft2, ifft2

    """

    # input must be complex!  Not exceptions
    if a.dtype != np.complex128 and a.dtype != np.complex64:
        raise ValueError('prop_fftw: Unsupported data type.  Must be complex64 or complex128.')


    # Create the description handle
    Desc_Handle = _ctypes.c_void_p(0)
    dims = (_ctypes.c_int64*2)(*a.shape)
   
    if a.dtype == np.complex64:
        mkl.DftiCreateDescriptor(_ctypes.byref(Desc_Handle), DFTI_SINGLE, DFTI_COMPLEX, _ctypes.c_int(2), dims)
    elif a.dtype == np.complex128:
        mkl.DftiCreateDescriptor(_ctypes.byref(Desc_Handle), DFTI_DOUBLE, DFTI_COMPLEX, _ctypes.c_int(2), dims)


    # Set normalization factor
    if norm == 'ortho':
        if a.dtype == np.complex64:
            scale = _ctypes.c_float(1.0 / np.sqrt(np.prod(a.shape)))
        else:
            scale = _ctypes.c_double(1.0 / np.sqrt(np.prod(a.shape)))
        mkl.DftiSetValue(Desc_Handle, DFTI_FORWARD_SCALE, scale)
        mkl.DftiSetValue(Desc_Handle, DFTI_BACKWARD_SCALE, scale)
    elif norm is None:
        if a.dtype == np.complex64:
            scale = _ctypes.c_float(1.0 / np.prod(a.shape))
        else:
            scale = _ctypes.c_double(1.0 / np.prod(a.shape))
        mkl.DftiSetValue(Desc_Handle, DFTI_BACKWARD_SCALE, scale)


    # Set input strides if necessary
    if not a.flags['C_CONTIGUOUS']:
        in_strides = (_ctypes.c_int*3)(0, a.strides[0] // a.itemsize, a.strides[1] // a.itemsize)
        mkl.DftiSetValue(Desc_Handle, DFTI_INPUT_STRIDES, _ctypes.byref(in_strides))

    if direction == 'forward':
        fft_func = mkl.DftiComputeForward
    elif direction == 'backward':
        fft_func = mkl.DftiComputeBackward
    else:
        assert False

    mkl.DftiSetValue( Desc_Handle, DFTI_THREAD_LIMIT, _ctypes.c_int(fft_nthreads) )

    # In-place FFT
    mkl.DftiCommitDescriptor( Desc_Handle )
    fft_func( Desc_Handle, a.ctypes.data_as(_ctypes.c_void_p) )
    mkl.DftiFreeDescriptor( _ctypes.byref(Desc_Handle) )

    return 

def mkl_fft2(a, norm=None, direction='forward', out=None):
    r"""Forward/backward 2D single- or double-precision FFT.

    Uses the Intel MKL libraries distributed with Enthought Python.
    Normalisation is different from Numpy!
    By default, allocates new memory like 'a' for output data.
    Returns the array containing output data.

    See Also
    --------
    fft2, ifft2

    """

    # convert input to complex data type if real (also memory copy)
    if a.dtype != np.complex128 and a.dtype != np.complex64:
        if a.dtype == np.int64 or a.dtype == np.uint64 or a.dtype == np.float64:
            a = np.array(a, dtype=np.complex128)
        else:
            a = np.array(a, dtype=np.complex64)

    # Configure in-place vs out-of-place
    inplace = False
    if out is a:
        inplace = True
    elif out is not None:
        assert out.dtype == a.dtype
        assert a.shape == out.shape
        assert not np.may_share_memory(a, out)
    else:
        out = np.empty_like(a)

    # Create the description handle
    Desc_Handle = _ctypes.c_void_p(0)
    dims = (_ctypes.c_long*2)(*a.shape)
   
    if a.dtype == np.complex64:
        mkl.DftiCreateDescriptor(_ctypes.byref(Desc_Handle), DFTI_SINGLE, DFTI_COMPLEX, _ctypes.c_int(2), dims)
    elif a.dtype == np.complex128:
        mkl.DftiCreateDescriptor(_ctypes.byref(Desc_Handle), DFTI_DOUBLE, DFTI_COMPLEX, _ctypes.c_int(2), dims)


    # Set normalization factor
    if norm == 'ortho':
        scale = _ctypes.c_double(1.0 / np.sqrt(np.prod(a.shape)))
        mkl.DftiSetValue(Desc_Handle, DFTI_FORWARD_SCALE, scale)
        mkl.DftiSetValue(Desc_Handle, DFTI_BACKWARD_SCALE, scale)

    elif norm is None:
        scale = _ctypes.c_double(1.0 / np.prod(a.shape))
        mkl.DftiSetValue(Desc_Handle, DFTI_FORWARD_SCALE, _ctypes.c_double(1.0))
        mkl.DftiSetValue(Desc_Handle, DFTI_BACKWARD_SCALE, scale)

        scale = _ctypes.c_float(0.)
        mkl.DftiGetValue(Desc_Handle, DFTI_BACKWARD_SCALE, _ctypes.byref(scale))


    # Set input strides if necessary
    if not a.flags['C_CONTIGUOUS']:
        in_strides = (_ctypes.c_long*3)(0, a.strides[0] // a.itemsize, a.strides[1] // a.itemsize)
        mkl.DftiSetValue(Desc_Handle, DFTI_INPUT_STRIDES, _ctypes.byref(in_strides))

    if direction == 'forward':
        fft_func = mkl.DftiComputeForward
    elif direction == 'backward':
        fft_func = mkl.DftiComputeBackward
    else:
        assert False

    if inplace:
        # In-place FFT
        mkl.DftiCommitDescriptor(Desc_Handle)
        fft_func(Desc_Handle, a.ctypes.data_as(_ctypes.c_void_p) )

    else:
        # Not-in-place FFT
        mkl.DftiSetValue(Desc_Handle, DFTI_PLACEMENT, DFTI_NOT_INPLACE)

        # Set output strides if necessary
        if not out.flags['C_CONTIGUOUS']:
            out_strides = (_ctypes.c_long*3)(0, out.strides[0] // out.itemsize, out.strides[1] // out.itemsize)
            mkl.DftiSetValue(Desc_Handle, DFTI_OUTPUT_STRIDES, _ctypes.byref(out_strides))

        mkl.DftiCommitDescriptor(Desc_Handle)
        fft_func(Desc_Handle, a.ctypes.data_as(_ctypes.c_void_p), out.ctypes.data_as(_ctypes.c_void_p) )

    mkl.DftiFreeDescriptor(_ctypes.byref(Desc_Handle))

    return out

def cce2full(A):

    # Assume all square for now

    N = A.shape
    N_half = N[0]//2 + 1
    out = np.empty((A.shape[0], A.shape[0]), dtype=A.dtype)
    out[:, :N_half] = A

    out[1:, N_half:] = np.rot90(A[1:, 1:-1], 2).conj()

    # Complete the first row
    out[0, N_half:] = A[0, -2:0:-1].conj()

    return out

def mkl_rfft2(a, norm=None, direction='forward', out=None):
    r"""Forward/backward single- or double-precision real-complex 2D FFT.

    For more details:

    See Also
    --------
    rfft2, irfft2

    """

    assert (a.dtype == np.float32) or (a.dtype == np.float64)

    out_type = np.complex128
    if a.dtype == np.float32:
        out_type = np.complex64

    n = a.shape[1]

    # Allocate memory if needed
    if out is not None:
        assert out.dtype == out_type
        assert out.shape[1] == n // 2 + 1
        assert not np.may_share_memory(a, out)
    else:
        size = list(a.shape)
        size[1] = n // 2 + 1
        out = np.empty(size, dtype=out_type)

    # Create the description handle
    Desc_Handle = _ctypes.c_void_p(0)
    dims = (_ctypes.c_long*2)(*a.shape)
   
    if a.dtype == np.float32:
        mkl.DftiCreateDescriptor(_ctypes.byref(Desc_Handle), DFTI_SINGLE, DFTI_REAL, _ctypes.c_int(2), dims)
    elif a.dtype == np.float64:
        mkl.DftiCreateDescriptor(_ctypes.byref(Desc_Handle), DFTI_DOUBLE, DFTI_REAL, _ctypes.c_int(2), dims)

    # Set the storage type
    mkl.DftiSetValue(Desc_Handle, DFTI_CONJUGATE_EVEN_STORAGE, DFTI_COMPLEX_COMPLEX)

    # Set normalization factor
    if norm == 'ortho':
        if a.dtype == np.float32:
            scale = _ctypes.c_float(1.0 / np.sqrt(np.prod(a.shape)))
        else:
            scale = _ctypes.c_double(1.0 / np.sqrt(np.prod(a.shape)))
        
        mkl.DftiSetValue(Desc_Handle, DFTI_FORWARD_SCALE, scale)
        mkl.DftiSetValue(Desc_Handle, DFTI_BACKWARD_SCALE, scale)
    elif norm is None:
        if a.dtype == np.float64:
            scale = _ctypes.c_float(1.0 / np.prod(a.shape))
        else:
            scale = _ctypes.c_double(1.0 / np.prod(a.shape))
        
        mkl.DftiSetValue(Desc_Handle, DFTI_BACKWARD_SCALE, scale)
    
    # For strides, the C type used *must* be long
    in_strides = (_ctypes.c_long*3)(0, a.strides[0] // a.itemsize, a.strides[1] // a.itemsize)
    out_strides = (_ctypes.c_long*3)(0, out.strides[0] // out.itemsize, out.strides[1] // out.itemsize)
    
    # mkl.DftiSetValue(Desc_Handle, DFTI_INPUT_STRIDES, _ctypes.byref(in_strides))
    mkl.DftiSetValue(Desc_Handle, DFTI_OUTPUT_STRIDES, _ctypes.byref(out_strides))

    if direction == 'forward':
        fft_func = mkl.DftiComputeForward
    elif direction == 'backward':
        fft_func = mkl.DftiComputeBackward
    else:
        assert False

    # Not-in-place FFT
    mkl.DftiSetValue(Desc_Handle, DFTI_PLACEMENT, DFTI_NOT_INPLACE)

    # Set output strides if necessary
    if not out.flags['C_CONTIGUOUS']:
        out_strides = (_ctypes.c_int*3)(0, out.strides[0] // out.itemsize, out.strides[1] // out.itemsize)
        mkl.DftiSetValue(Desc_Handle, DFTI_OUTPUT_STRIDES, _ctypes.byref(out_strides))

    mkl.DftiCommitDescriptor(Desc_Handle)
    fft_func(Desc_Handle, a.ctypes.data_as(_ctypes.c_void_p), out.ctypes.data_as(_ctypes.c_void_p) )

    mkl.DftiFreeDescriptor(_ctypes.byref(Desc_Handle))

    return out


def rfft(a, n=None, axis=-1, norm=None, out=None, scrambled=False):
    r"""Computes the forward real-complex FFT using Intel's MKL routines.

    Faster than mkl_fft.fft for real arrays. 

    Parameters 
    ----------
    a : ndarray
        Input array to transform. It must be real.
    n : int
        Size of the transform.
    axis : int
        Axis along which the transform is computed (default is -1, summation
        over the last axis).
    norm : {None, 'ortho'}
        Normalization of the transform. None (default) is same as numpy;
        'ortho' gives an orthogonal (norm-preserving) transform.
    out : ndarray
        Points to the output array. Used when the array is preallocated or an
        in-place transform is desired. Default is None, meaning that the
        memory is allocated for the output array of the same shape as a.
    scrambled: bool, optional (default False)
        Allows the output of the FFT to be out of order if set to true. This
        can sometimes lead to better performance.

    Returns
    -------
        The transformed output array.

    """
    return mkl_rfft(a, n=n, axis=axis, norm=norm, direction='forward', out=out, scrambled=scrambled)

def irfft(a, n=None, axis=-1, norm=None, out=None, scrambled=False):
    r"""Computes the inverse complex-real FFT using Intel's MKL routines.

    Faster than mkl_fft.ifft for conjugate-even arrays. 

    Parameters
    ----------
    a : ndarray
        Input array to transform. It should be stored in the conjugate-even
        format (i.e. like output of rfft).
    n : int
        Size of the transform.
    axis : int
        Axis along which the transform is computed (default is -1, summation
        over the last axis).
    norm : {None, 'ortho'}
        Normalization of the transform. None (default) is same as numpy;
        'ortho' gives an orthogonal (norm-preserving) transform.
    out : ndarray
        Points to the output array. Used when the array is preallocated or an
        in-place transform is desired. Default is None, meaning that the
        memory is allocated for the output array of the same shape as a.
    scrambled: bool, optional (default False)
        Allows the input of the iFFT to be out of order if set to true. This
        can sometimes lead to better performance.

    Returns
    -------
        The transformed output array.

    """
    return mkl_rfft(a, n=n, axis=axis, norm=norm, direction='backward', out=out, scrambled=scrambled)

def fft(a, n=None, axis=-1, norm=None, out=None, scrambled=False):
    r"""Computes the forward FFT using Intel's MKL routines.

    Parameters
    ----------
    a : ndarray
        Input array to transform.
    n : int
        Size of the transform.
    axis : int
        Axis along which the transform is computed (default is -1, summation
        over the last axis).
    norm : {None, 'ortho'}
        Normalization of the transform. None (default) is same as numpy;
        'ortho' gives an orthogonal (norm-preserving) transform.
    out : ndarray
        Points to the output array. Used when the array is preallocated or an
        in-place transform is desired. Default is None, meaning that the
        memory is allocated for the output array of the same shape as a.
    scrambled: bool, optional (default False)
        Allows the output of the FFT to be out of order if set to true. This
        can sometimes lead to better performance.

    Returns
    -------
        The transformed output array.

    """
    return mkl_fft(a, n=n, axis=axis, norm=norm, direction='forward', out=out, scrambled=scrambled)

def ifft(a, n=None, axis=-1, norm=None, out=None, scrambled=False):
    r"""Computes the inverse FFT using Intel's MKL routines.

    Parameters
    ----------
    a : ndarray
        Input array to transform.
    n : int
        Size of the transform.
    axis : int
        Axis along which the transform is computed (default is -1, summation
        over the last axis).
    norm : {None, 'ortho'}
        Normalization of the transform. None (default) is same as numpy;
        'ortho' gives an orthogonal (norm-preserving) transform.
    out : ndarray
        Points to the output array. Used when the array is preallocated or an
        in-place transform is desired. Default is None, meaning that the
        memory is allocated for the output array of the same shape as a.
    scrambled: bool, optional (default False)
        Allows the input of the iFFT to be out of order if set to true. This
        can sometimes lead to better performance.

    Returns
    -------
        The transformed output array.

    """
    return mkl_fft(a, n=n, axis=axis, norm=norm, direction='backward', out=out, scrambled=scrambled)

def fft2(a, norm=None, out=None):
    r"""Computes the forward 2D FFT using Intel's MKL routines.

    Parameters
    ----------
    a : ndarray
        Input array to transform.
    norm : {None, 'ortho'}
        Normalization of the transform. None (default) is same as numpy;
        'ortho' gives an orthogonal (norm-preserving) transform.
    out : ndarray
        Points to the output array. Used when the array is preallocated or an
        in-place transform is desired. Default is None, meaning that the
        memory is allocated for the output array of the same shape as a.

    Returns
    -------
        The transformed output array.

    """
    proper_fft2(a, norm=norm, direction='forward')
    return fftshift(a)
    #return mkl_fft2(a, norm=norm, direction='forward', out=out)

def ifft2(a, norm=None, out=None):
    r"""Computes the inverse 2D FFT using Intel's MKL routines.

    Parameters
    ----------
    a : ndarray
        Input array to transform.
    norm : {None, 'ortho'}
        Normalization of the transform. None (default) is same as numpy;
        'ortho' gives an orthogonal (norm-preserving) transform.
    out : ndarray
        Points to the output array. Used when the array is preallocated or an
        in-place transform is desired. Default is None, meaning that the
        memory is allocated for the output array of the same shape as a.

    Returns
    -------
        The transformed output array.

    """

    proper_fft2(a, norm=norm, direction='backward')
    return fftshift(a)
    #return mkl_fft2(a, norm=norm, direction='backward', out=out)


def rfft2(a, norm=None, out=None):
    r"""Computes the forward real -> complex conjugate-even 2D FFT using
    Intel's MKL routines.

    Faster than mkl_fft.fft2 for real arrays.

    Parameters
    ----------
    a : ndarray
        Input array to transform. It must be real.
    norm : {None, 'ortho'}
        Normalization of the transform. None (default) is same as numpy;
        'ortho' gives an orthogonal (norm-preserving) transform.
    out : ndarray
        Points to the output array. Used when the array is preallocated or an
        in-place transform is desired. Default is None, meaning that the
        memory is allocated for the output array of the same shape as a.

    Returns
    -------
        The transformed output array.

    """

    return mkl_rfft2(a, norm=None, direction='forward', out=None)


def irfft2(a, norm=None, out=None):
    r"""Computes the forward conjugate-even -> real 2D FFT using Intel's MKL
    routines.

    Faster than mkl_fft.ifft2 for conjugate-even arrays.

    Parameters
    ----------
    a : ndarray
        Input array to transform. It should be stored in the conjugate-even
        format (i.e. like output of rfft2).
    norm : {None, 'ortho'}
        Normalization of the transform. None (default) is same as numpy;
        'ortho' gives an orthogonal (norm-preserving) transform.
    out : ndarray
        Points to the output array. Used when the array is preallocated or an
        in-place transform is desired. Default is None, meaning that the
        memory is allocated for the output array of the same shape as a.

    Returns
    -------
        The transformed output array.

    """

    return mkl_rfft2(a, norm=None, direction='backward', out=None)

def fftshift(x, additional_shift=None, axes=None):
    """
    Shift the zero-frequency component to the center of the spectrum, or with
    some additional offset from the center.

    This is a more generic fork of `~numpy.fft.fftshift`, which doesn't support
    additional shifts.


    Parameters
    ----------
    x : array_like
        Input array.
    additional_shift : list of length ``M``
        Desired additional shifts in ``x`` and ``y`` directions respectively
    axes : int or shape tuple, optional
        Axes over which to shift.  Default is None, which shifts all axes.

    Returns
    -------
    y : `~numpy.ndarray`
        The shifted array.
    """
    tmp = np.asarray(x)
    ndim = len(tmp.shape)
    if axes is None:
        axes = list(range(ndim))
    elif isinstance(axes, integer_types):
        axes = (axes,)

    # If no additional shift is supplied, reproduce `numpy.fft.fftshift` result
    if additional_shift is None:
        additional_shift = [0, 0]

    y = tmp
    for k, extra_shift in zip(axes, additional_shift):
        n = tmp.shape[k]
        if (n+1)//2 - extra_shift < n:
            p2 = (n+1)//2 - extra_shift
        else:
            p2 = abs(extra_shift) - (n+1)//2
        mylist = np.concatenate((np.arange(p2, n), np.arange(0, p2)))
        y = np.take(y, mylist, k)
    return y


if __name__ == "__main__":

    import time

    n_iter = 200
    N = 256

    np.seterr(all='raise')

    algos = {
            'Numpy fft2 complex128': {'transform': np.fft.fft2, 'dtype': np.complex128},
            'MKL fft2 complex128': {'transform': fft2, 'dtype': np.complex128},
            'Numpy fft2 complex64': {'transform': np.fft.fft2, 'dtype': np.complex64},
            'MKL fft2 complex64': {'transform': fft2, 'dtype': np.complex64},
            'Numpy fft complex128': {'transform': np.fft.fft, 'dtype': np.complex128},
            'MKL fft complex128': {'transform': fft, 'dtype': np.complex128},
            'Numpy fft complex64': {'transform': np.fft.fft, 'dtype': np.complex64},
            'MKL fft complex64': {'transform': fft, 'dtype': np.complex64},
            'Numpy rfft float64': {'transform': np.fft.rfft, 'dtype': np.float64},
            'MKL rfft float64': {'transform': rfft, 'dtype': np.float64},
            'Numpy rfft float32': {'transform': np.fft.rfft, 'dtype': np.float32},
            'MKL rfft float32': {'transform': rfft, 'dtype': np.float32},
            }

    for algo in algos.keys():

        A = algos[algo]['dtype'](np.random.randn(N, N))
        #C = np.zeros((N, N), dtype='complex128')
        start_time = time.time()
        for i in range(n_iter):
            algos[algo]['transform'](A)
        total = time.time() - start_time
        print(algo,":")
        print("--- %s seconds ---" % total)





"""Estimation functions for WFSC."""

import numpy as np
import multiprocessing
# from astropy.io import fits
import matplotlib.pyplot as plt
import falco
from . import check

def perfect(mp):
    """
    Return the perfect-knowledge E-field from the full model.
    
    Optionally add Zernikes at the input pupil.

    Parameters
    ----------
    mp : ModelParameters
        Structure containing optical model parameters
        
    Returns
    -------
    Emat : numpy ndarray
        2-D array with the vectorized, complex E-field of the dark hole pixels
        for each mode included in the control Jacobian.
    """
    if type(mp) is not falco.config.ModelParameters:
        raise TypeError('Input "mp" must be of type ModelParameters')
    
    if mp.flagMultiproc:
        
        Emat = np.zeros((mp.Fend.corr.Npix, mp.jac.Nmode), dtype=complex)
        
        # Loop over all modes and wavelengths
        inds_list = [(x, y) for x in range(mp.jac.Nmode) for y in range(mp.Nwpsbp)]
        Nvals = mp.jac.Nmode*mp.Nwpsbp

        pool = multiprocessing.Pool(processes=mp.Nthreads)
        resultsRaw = [pool.apply_async(_est_perfect_Efield_with_Zernikes_in_parallel,
                                       args=(mp, ilist, inds_list)) for ilist in range(Nvals)]
        results = [p.get() for p in resultsRaw]  # All the images in a list
        pool.close()
        pool.join()
        
        # Re-order for easier indexing
        Ecube = np.zeros((mp.Fend.corr.Npix, mp.jac.Nmode, mp.Nwpsbp), dtype=complex)
        for iv in range(Nvals):
            im = inds_list[iv][0]  # Index of the Jacobian mode
            wi = inds_list[iv][1]   # Index of the wavelength in the sub-bandpass
            Ecube[:, im, wi] = results[iv]
        Emat = np.mean(Ecube, axis=2)  # Average over wavelengths in the subband
  
#        EmatAll = np.zeros((mp.Fend.corr.Npix, Nval))
#        for iv in range(Nval):
#            EmatAll[:, iv] = results[iv]
#
#        counter = 0;
#        for im=1:mp.jac.Nmode
#            EsbpMean = 0;
#            for wi=1:mp.Nwpsbp
#                counter = counter + 1;
#                EsbpMean = EsbpMean + EmatAll(:,counter) * \
#                     mp.full.lambda_weights(wi);
#            end
#            Emat(:,im) = EsbpMean;
#        end
    
    else:
    
        Emat = np.zeros((mp.Fend.corr.Npix, mp.jac.Nmode), dtype=complex)
        modvar = falco.config.Object()
        
        for im in range(mp.jac.Nmode):
            modvar.sbpIndex = mp.jac.sbp_inds[im]
            modvar.zernIndex = mp.jac.zern_inds[im]
            modvar.whichSource = 'star'
            
            # Take the mean over the wavelengths within the sub-bandpass
            EmatSbp = np.zeros((mp.Fend.corr.Npix, mp.Nwpsbp), dtype=complex)
            for wi in range(mp.Nwpsbp):
                modvar.wpsbpIndex = wi
                E2D = falco.model.full(mp, modvar)
                # Actual field in estimation area. Apply spectral weight
                # within the sub-bandpass
                EmatSbp[:, wi] = mp.full.lambda_weights[wi] * \
                    E2D[mp.Fend.corr.maskBool]
            Emat[:, im] = np.sum(EmatSbp, axis=1)
            
    return Emat
    

# Extra function needed to use parfor (because parfor can have only a
# single changing input argument).
def _est_perfect_Efield_with_Zernikes_in_parallel(mp, ilist, inds_list):

    im = inds_list[ilist][0]  # Index of the Jacobian mode
    wi = inds_list[ilist][1]   # Index of the wavelength in the sub-bandpass
    
    modvar = falco.config.Object()
    modvar.sbpIndex = mp.jac.sbp_inds[im]
    modvar.zernIndex = mp.jac.zern_inds[im]
    modvar.wpsbpIndex = wi
    modvar.whichSource = 'star'
    
    E2D = falco.model.full(mp, modvar)

    # Actual field in estimation area. Don't apply spectral weight here.
    return E2D[mp.Fend.corr.maskBool]


def pairwise_probing(mp, ev, jacStruct=np.array([])):
    """
    Estimate the dark hole E-field with pair-wise probing.

    Parameters
    ----------
    mp : falco.config.ModelParameter
        Object containing all model parameters.
    ev : falco.config.Object()
    jacStruct : array_like, optional
        Array containing the control Jacobian. Default is an empty array.

    Returns
    -------
    None
        Outputs are included in the object ev.
        
    """
    # "ev" is passed in only for the Kalman filter. Reset it for the batch
    # process to avoid accidentally using old data.
    # if 'pwp-bp' == mp.estimator.lower():
    #     ev = falco.config.Object()
    
    # Select number of actuators across based on chosen DM for the probing
    if mp.est.probe.whichDM == 1:
        Nact = mp.dm1.Nact
    elif mp.est.probe.whichDM == 2:
        Nact = mp.dm2.Nact
    else:
        raise ValueError('mp.est.probe.whichDM must equal 1 or 2.')
    
    # Store the initial DM commands
    if np.any(mp.dm_ind == 1):
        DM1Vnom = mp.dm1.V
    
    if np.any(mp.dm_ind == 2):
        DM2Vnom = mp.dm2.V
    else:
        DM2Vnom = np.zeros_like(mp.dm1.V)
    
    # Definitions:
    Npairs = mp.est.probe.Npairs  # Number of image PAIRS
    ev.Icube = np.zeros((mp.Fend.Neta, mp.Fend.Nxi, 1+2*Npairs))
    if np.any(mp.dm_ind == 1):
        ev.Vcube1 = np.zeros((mp.dm1.Nact, mp.dm1.Nact, 1+2*Npairs))
    if np.any(mp.dm_ind == 2):
        ev.Vcube2 = np.zeros((mp.dm2.Nact, mp.dm2.Nact, 1+2*Npairs))
    
    # Generate evenly spaced probes along the complex unit circle
    # NOTE: Nprobes=Npairs*2
    probePhaseVec = np.array([0, Npairs])
    for k in range(Npairs-1):
        probePhaseVec = np.append(probePhaseVec, probePhaseVec[-1]-(Npairs-1))
        probePhaseVec = np.append(probePhaseVec, probePhaseVec[-1]+Npairs)
    probePhaseVec = probePhaseVec*np.pi/(Npairs)
    
    badAxisVec = ''
    if mp.est.probe.axis.lower() == 'y':
        for _iter in range(2*Npairs):
            badAxisVec += 'y'
    elif mp.est.probe.axis.lower() == 'x':
        for _iter in range(2*Npairs):
            badAxisVec += 'x'
    elif mp.est.probe.axis.lower() in ('alt', 'xy', 'alternate'):
        for iPair in range(2*Npairs):
            if (iPair+1) % 4 == 1 or (iPair+1) % 4 == 2:
                badAxisVec += 'x'
            elif (iPair+1) % 4 == 3 or (iPair+1) % 4 == 0:
                badAxisVec += 'y'
    elif mp.est.probe.axis.lower() in ('m', 'multi'):
        for _iter in range(2*Npairs):
            badAxisVec += 'm'
    else:
        raise ValueError('Incorrect value for mp.est.probe.axis')
    
    # Initialize output arrays
    ev.Eest = np.zeros((mp.Fend.corr.Npix, mp.Nsbp), dtype=complex)
    ev.IincoEst = np.zeros((mp.Fend.corr.Npix, mp.Nsbp))
    ev.I0mean = 0
    ev.IprobedMean = 0
    
    # Get images and perform estimates in each sub-bandpass
    print('Estimating electric field with batch process estimation ...')
    
    for si in range(mp.Nsbp):
        print('Wavelength: %u/%u ... ' % (si, mp.Nsbp-1))
    
        # Valid for all calls to model_compact.m:
        modvar = falco.config.Object()  # Initialize the new structure
        modvar.sbpIndex = si
        modvar.whichSource = 'star'
    
        # Measure current contrast level average
        # Reset DM commands to the unprobed state:
        mp.dm1.V = DM1Vnom
        mp.dm2.V = DM2Vnom
        # Separate out image values at DH pixels and delta DM voltage settings
        Iplus = np.zeros((mp.Fend.corr.Npix, Npairs))
        Iminus = np.zeros((mp.Fend.corr.Npix, Npairs))
        DM1Vplus = np.zeros((Nact, Nact, Npairs))
        DM1Vminus = np.zeros((Nact, Nact, Npairs))
        DM2Vplus = np.zeros((Nact, Nact, Npairs))
        DM2Vminus = np.zeros((Nact, Nact, Npairs))
    
        # Compute probe shapes and take probed images:
    
        # Take initial, unprobed image (for unprobed DM settings).
        whichImg = 1
        I0 = falco.imaging.get_sbp_image(mp, si)
        I0vec = I0[mp.Fend.corr.maskBool]  # Vectorize the correction region
        ev.I0mean = ev.I0mean+I0/mp.Nsbp  # Getting Inorm for whole bandpass
    
        # Store values for first image and its DM commands
        ev.Icube[:, :, whichImg] = I0
        if np.any(mp.dm_ind == 1):
            ev.Vcube1[:, :, whichImg] = mp.dm1.V
        if np.any(mp.dm_ind == 2):
            ev.Vcube2[:, :, whichImg] = mp.dm2.V
    
        # Compute the average Inorm in the scoring and correction regions
        ev.InormScore = np.mean(I0[mp.Fend.score.maskBool])
        ev.InormCorr = np.mean(I0[mp.Fend.corr.maskBool])
        print('Measured unprobed Inorm (Corr / Score): %.2e \t%.2e \n' %
              (ev.InormCorr, ev.InormScore))
    
        # Set (approximate) probe intensity based on current measured Inorm
        if mp.flagFiber:
            ev.InormProbeMax = 1e-5
            InormProbe = np.min([np.sqrt(np.max(I0)*1e-8), ev.InormProbeMax])
        else:
            ev.InormProbeMax = 1e-4
            InormProbe = np.min([np.sqrt(np.max(I0vec)*1e-5),
                                 ev.InormProbeMax])
            # Change this to a high percentile value (e.g., 90%) instead of the
            # max to avoid being tricked by noise
        print('Chosen probe intensity: %.2e' % InormProbe)
    
        # Perform the probing
        iOdd = 0  # Initialize index counters
        iEven = 0
        for iProbe in range(2*Npairs):

            # Generate the command map for the probe
            probeCmd = gen_pairwise_probe(mp, InormProbe,
                                probePhaseVec[iProbe], badAxisVec[iProbe])
    
            # Select which DM to use for probing. Allocate probe to that DM
            if mp.est.probe.whichDM == 1:
                dDM1Vprobe = probeCmd/mp.dm1.VtoH  # Now in volts
                dDM2Vprobe = 0
            elif mp.est.probe.whichDM == 2:
                dDM1Vprobe = 0
                dDM2Vprobe = probeCmd/mp.dm1.VtoH  # Now in volts
            else:
                raise ValueError('DM for probing must be 1 or 2.')
                
            if np.any(mp.dm_ind == 1):
                mp.dm1.V = DM1Vnom + dDM1Vprobe
            if np.any(mp.dm_ind == 2):
                mp.dm2.V = DM2Vnom + dDM2Vprobe
                    
            # Take probed image
            if mp.flagFiber:
                Im = falco.imaging.get_sbp_image_fiber(mp, si)
            else:
                Im = falco.imaging.get_sbp_image(mp, si)
            
            # plt.imshow(np.log10(Im)); plt.title('Probed Image %d' % iProbe); plt.colorbar(); plt.pause(1e-2);
            
            ImNonneg = Im
            ImNonneg[Im < 0] = 0
            whichImg = 1+iProbe  # Increment image counter
            # Inorm averaged over all the probed images
            ev.IprobedMean = ev.IprobedMean + \
                np.mean(Im[mp.Fend.corr.maskBool]) / (2*Npairs)
    
            # Store probed image and its DM settings
            ev.Icube[:, :, whichImg] = Im
            if np.any(mp.dm_ind == 1):
                ev.Vcube1[:, :, whichImg] = mp.dm1.V
            if np.any(mp.dm_ind == 2):
                ev.Vcube2[:, :, whichImg] = mp.dm2.V
    
            # Report results
            probeSign = '-+'
            print('Actual Probe %d%s Contrast is: %.2e' % (
                    np.floor(iProbe/2), probeSign[(iProbe+1) % 2],
            np.mean(Im[mp.Fend.corr.maskBool])))

            # Assign image to positive or negative probe collection:
            if (iProbe+1) % 2 == 1:  # Odd; for plus probes
                if np.any(mp.dm_ind == 1):
                    DM1Vplus[:, :, iOdd] = dDM1Vprobe + DM1Vnom
                if np.any(mp.dm_ind == 2):
                    DM2Vplus[:, :, iOdd] = dDM2Vprobe + DM2Vnom
                Iplus[:, iOdd] = Im[mp.Fend.corr.maskBool]
                iOdd += 1
            elif (iProbe+1) % 2 == 0:  # Even; for minus probes
                if np.any(mp.dm_ind == 1):
                    DM1Vminus[:, :, iEven] = dDM1Vprobe + DM1Vnom
                if np.any(mp.dm_ind == 2):
                    DM2Vminus[:, :, iEven] = dDM2Vprobe + DM2Vnom
                Iminus[:, iEven] = Im[mp.Fend.corr.maskBool]
                iEven += 1
    
        # Calculate probe amplitudes and measurement vector.
        # (Refer again to Give'on+ SPIE 2011 to undersand why.)
        ampSq = (Iplus+Iminus)/2 - np.tile(I0vec.reshape((-1, 1)), (1, Npairs))  # square of probe E-field amplitudes
        ampSq[ampSq < 0] = 0  # If probe amplitude is zero, set amp = 0
        amp = np.sqrt(ampSq)  # E-field amplitudes, dimensions: [mp.Fend.corr.Npix, Npairs]
        isnonzero = np.all(amp, 1)
        zAll = ((Iplus-Iminus)/4).T  # Measurement vector, dimensions: [Npairs,mp.Fend.corr.Npix]
        ampSq2Dcube = np.zeros((mp.Fend.Neta, mp.Fend.Nxi, mp.est.probe.Npairs))
        for iProbe in range(Npairs):  # Display the actual probe intensity
            ampSq2D = np.zeros((mp.Fend.Neta, mp.Fend.Nxi))
            ampSq2D[mp.Fend.corr.maskBool] = ampSq[:, iProbe]
            ampSq2Dcube[:, :, iProbe] = ampSq2D
            print('*** Mean measured Inorm for probe #%d  =\t%.3e' %
                  (iProbe, np.mean(ampSq2D[mp.Fend.corr.maskBool])))
    
        # Plot relevant data for all the probes
        # falco_plot_pairwi1se_probes(mp, ev,
        #         DM1Vplus-repmat(DM1Vnom, [1,1,size(DM1Vplus,3)]),
        #         ampSq2Dcube)
    
        # ################# Perform the estimation ############################
        
        # Use Jacobian for estimation. This is fully model-based if the
        # Jacobian is purely model-based, or it is better if the Jacobian is
        # adaptive based on empirical data.
        if mp.est.flagUseJac:
            
            dEplus = np.zeros_like(Iplus, dtype=complex)
            for iProbe in range(Npairs):
                if mp.est.probe.whichDM == 1:
                    dV = DM1Vplus[:, :, iProbe] - DM1Vnom
                    dEplus[:, iProbe] = np.squeeze(jacStruct.G1[:, :, si]) * \
                        dV[mp.dm1.act_ele]
                elif mp.est.probe.whichDM == 2:
                    dV = DM2Vplus[:, :, iProbe] - DM2Vnom
                    dEplus[:, iProbe] = np.squeeze(jacStruct.G2[:, :, si]) * \
                        dV[mp.dm2.act_ele]

        # Get the probe phase from the model and measure the probe amplitude
        else:

            # For unprobed field based on model:
            if np.any(mp.dm_ind == 1):
                mp.dm1.V = DM1Vnom
            if np.any(mp.dm_ind == 2):
                mp.dm2.V = DM2Vnom
            if mp.flagFiber:
                pass
                # [~, E0] = model_compact(mp, modvar)
            else:
                E0 = falco.model.compact(mp, modvar)

            E0vec = E0[mp.Fend.corr.maskBool]
    
            # For probed fields based on model:
            Eplus = np.zeros_like(Iplus, dtype=complex)
            Eminus = np.zeros_like(Iminus, dtype=complex)
            for iProbe in range(Npairs):
                # For plus probes:
                if np.any(mp.dm_ind == 1):
                    mp.dm1.V = np.squeeze(DM1Vplus[:, :, iProbe])
                if np.any(mp.dm_ind == 2):
                    mp.dm2.V = np.squeeze(DM2Vplus[:, :, iProbe])
                if(mp.flagFiber):
                    pass
                    # [~, Etemp] = model_compact(mp, modvar);
                else:
                    Etemp = falco.model.compact(mp, modvar)
                Eplus[:, iProbe] = Etemp[mp.Fend.corr.maskBool]
                
                # For minus probes:
                if np.any(mp.dm_ind == 1):
                    mp.dm1.V = np.squeeze(DM1Vminus[:, :, iProbe])
                if np.any(mp.dm_ind == 2):
                    mp.dm2.V = np.squeeze(DM2Vminus[:, :, iProbe])
                if mp.flagFiber:
                    pass
                    # [~, Etemp] = model_compact(mp, modvar)
                else:
                    Etemp = falco.model.compact(mp, modvar)
                Eminus[:, iProbe] = Etemp[mp.Fend.corr.maskBool]
    
            # Create delta E-fields for each probe image.
            # Then create Npairs phase angles.
            dEplus = Eplus - np.tile(E0vec.reshape((-1, 1)), (1, Npairs))
            dEminus = Eminus - np.tile(E0vec.reshape((-1, 1)), (1, Npairs))
            dphdm = np.zeros((mp.Fend.corr.Npix, Npairs))  # phases
            for iProbe in range(Npairs):
                dphdm[:, iProbe] = np.arctan2(
                    np.imag(dEplus[:, iProbe]) - np.imag(dEminus[:, iProbe]),
                    np.real(dEplus[:, iProbe]) - np.real(dEminus[:, iProbe]))
                    
        # Batch process the measurements to estimate the electric field in the
        # dark hole. Done pixel by pixel.
        
        if (mp.estimator.lower() == 'pwp-bp') or \
            (mp.estimator.lower() == 'pwp-kf' and ev.Itr < mp.est.ItrStartKF):
    
            Eest = np.zeros((mp.Fend.corr.Npix,), dtype=complex)
            zerosCounter = 0  # number of zeroed-out dark hole pixels
            for ipix in range(mp.Fend.corr.Npix):
                
                if mp.est.flagUseJac:
                    dE = dEplus[ipix, :].T
                    H = np.array([np.real(dE), np.imag(dE)])
                else:
                    H = np.zeros([Npairs, 2])  # Observation matrix
                    # Leave Eest for a pixel as zero if any probe amp is 0
                    if isnonzero[ipix] == 1:
                        for iProbe in range(Npairs):
                            H[iProbe, :] = amp[ipix, iProbe] * \
                                np.array([np.cos(dphdm[ipix, iProbe]),
                                          np.sin(dphdm[ipix, iProbe])])
                    else:
                        zerosCounter += 1
    
                Epix = np.linalg.pinv(H) @ zAll[:, ipix]  # Batch processing
                Eest[ipix] = Epix[0] + 1j*Epix[1]

            # If estimate is too bright, the estimate was probably bad.
            # !!!!!!!!!!!!!!BE VERY CAREFUL WITH THIS HARD-CODED VALUE!!!!!!!!!
            Eest[np.abs(Eest)**2 > 1e-2] = 0.0
            
            print('%d of %d pixels were given zero probe amplitude.' %
                    (zerosCounter, mp.Fend.corr.Npix))
        
            # Initialize the state and state covariance estimates for Kalman
            # filter. The state is the real and imag parts of the E-field.
            if mp.estimator.lower() == 'pwp-kf':
                # Re-organize the batch-processed E-field estimate into the 1st
                # state estimate for the Kalman filter
                xOld = np.zeros((2*mp.Fend.corr.Npix, 1))
                for ii in range(mp.Fend.corr.Npix):
                    xOld[2*(ii-1)+0:2*(ii-1)+1] = np.array([np.real(Eest[ii]),
                                                            np.imag(Eest[ii])])
                ev.xOld = xOld  # Save out for returning later
    
                # Initialize the state covariance matrix
                # (2x2 for each dark hole pixel)
                ev.Pold_KF_array = np.tile(mp.est.Pcoef0*np.eye(2),
                                           (mp.Fend.corr.Npix, 1, mp.Nsbp))
    
        # %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        # Begin Kalman Filter Update
        # %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        #
        # To be completed later...
        #
        # %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        # End Kalman Filter Update
        # %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        
        # Save out the estimates
        ev.Eest[:, si] = Eest
        Iest = np.abs(Eest)**2
        ev.IincoEst[:, si] = I0vec - Iest  # incoherent light
    
    # Other data to save out
    ev.ampSqMean = np.mean(ampSq)  # Mean probe intensity
    ev.ampNorm = amp/np.sqrt(InormProbe)  # Normalized probe amplitude maps
    
    # Calculate the mean normalized intensity over the whole dark hole at all
    # wavelengths.
    ev.InormEst = np.mean(Iest)
    
    # Reset DM commands to their values before probing
    if np.any(mp.dm_ind == 1):
        mp.dm1.V = DM1Vnom
    
    if np.any(mp.dm_ind == 2):
        mp.dm2.V = DM2Vnom
    
    print('Completed pairwise probing estimation.')
 
    pass
    # return None


def gen_pairwise_probe(mp, InormDes, psi, badAxis):
    """
    Generate delta DM commands that probe the dark hole.
    
    Parameters
    ----------
    mp : falco.config.ModelParameter
        Object containing all model parameters.
    InormDes : float
        Desired normalized intensity of the probes in the image.
    psi : float
        phase angle of the sinusoidal part of the probe. Units of radians.
    badAxis : TYPE
        DESCRIPTION.

    Returns
    -------
    probeCmd : array_like
         Nact x Nact array of delta DM actuator commands to make a probe.
    """
    check.real_positive_scalar(InormDes, 'InormDes', ValueError)
    check.real_scalar(psi, 'psi', TypeError)
    if badAxis.lower() not in ('x', 'y', 'm'):
        raise ValueError('Invalid value for badAxis.')
        
    # Number of actuators across DM surface
    # (independent of beam diameter for time being)
    if mp.est.probe.whichDM == 1:
        Nact = mp.dm1.Nact
        dm = mp.dm1
    elif mp.est.probe.whichDM == 2:
        Nact = mp.dm2.Nact
        dm = mp.dm2
    
    # Coordinates in actuator space
    xs = np.arange(-(Nact-1)/2, (Nact+1)/2)/Nact - \
        np.round(mp.est.probe.offsetX)/Nact
    ys = np.arange(-(Nact-1)/2, (Nact+1)/2)/Nact - \
        np.round(mp.est.probe.offsetY)/Nact
    [XS, YS] = np.meshgrid(xs, ys)
    
    # Restrict the probing region if it is not possible to achieve
    if mp.est.probe.radius > Nact/2.0:
        mp.est.probe.radius = Nact/2.0
    
    # Generate the DM command for the probe
    magn = 4*np.pi*mp.lambda0*np.sqrt(InormDes)  # surface height to get desired intensity [meters]
    if badAxis.lower() == 'y':
        mX = mp.est.probe.radius
        mY = 2*mp.est.probe.radius
        omegaX = mp.est.probe.radius/2
        probeSurf = magn*np.sinc(mX*XS)*np.sinc(mY*YS)*np.cos(2*np.pi*omegaX*XS + psi)

    elif badAxis.lower() == 'x':
        mX = 2*mp.est.probe.radius
        mY = mp.est.probe.radius
        omegaY = mp.est.probe.radius/2
        probeSurf = magn*np.sinc(mX*XS)*np.sinc(mY*YS)*np.cos(2*np.pi*omegaY*YS + psi)
    
    elif badAxis.lower() == 'm':
        omegaX = mp.est.probe.Xloc/2
        omegaY = mp.est.probe.Yloc/2
        probeSurf = np.zeros_like(XS)
        for i in range(mp.Fend.Nfiber):
            probeSurf = probeSurf + \
                magn*np.sin(2*np.pi*omegaX(i)*XS + 2*np.pi*omegaY(i)*YS + psi)

    # Option to use just the sincs for a zero phase shift. This avoids the
    # phase discontinuity along one axis (for this probe only!).
    if psi == 0:
        m = 2*mp.est.probe.radius
        probeSurf = magn*np.sinc(m*XS)*np.sinc(m*YS)
    
    probeCmd = falco.dm.fit_surf_to_act(dm, probeSurf)
    
    # Scale the probe amplitude empirically if needed
    probeCmd = mp.est.probe.gainFudge*probeCmd
    
    return probeCmd
